(in-package "NEWCMP")


;;;;
;;;; Call Graphs
;;;;

(defstruct (callgr (:constructor make-callgr (arg-refs-p)))
  functions
  (callee-table (make-hash-table))
  (caller-table (make-hash-table))
  arg-refs-p)

(defun add-callgr-function (fun gr)
  (pushnew fun (callgr-functions gr)))

(defun add-callgr-call (caller callee gr)
  (push callee (gethash caller (callgr-callee-table gr)))
  (push caller (gethash callee (callgr-caller-table gr))))

(defun callgr-callees (f gr)
  (gethash f (callgr-callee-table gr)))

(defun callgr-callers (f gr)
  (gethash f (callgr-caller-table gr)))

(defun show-callgr (gr)
  (let ((val nil))
    (maphash #'(lambda (x y)
                 (push (cons (func-name x) (mapcar #'func-name y)) val))
             (callgr-callee-table gr))
    val))

(defun compute-callgr (e &optional arg-refs-p)
  (let ((gr (make-callgr arg-refs-p)))
    (fill-callgr e e nil gr)
    gr))

(defun fill-callgr (e fun env gr)
  (cond
   ((func-p e) (add-callgr-function e gr)
               (fill-callgr (func-body e) e env gr))
   ((letrec-p e) (let* ((args (letrec-args e))
                        (vars (letrec-vars e))
                        (nenv (append (mapcar #'cons vars args) env)))
                   (dolist (a args) (fill-callgr a a nenv gr))
                   (fill-callgr (letrec-body e) fun nenv gr)))
   ((call-p e) (let ((lfun (cdr (assoc (call-fun e) env)))
                     (form (call-form e)))
                 (when lfun (add-callgr-call fun lfun gr))
                 (when (callgr-arg-refs-p gr)
                       (dolist (a (cexpr-args e))
                         (let ((lfun (cdr (assoc a env))))
                           (when lfun (add-callgr-call fun lfun gr)))))
                 (when form (fill-callgr form fun env gr))))
   (t (when (callgr-arg-refs-p gr)
            (dolist (a (cexpr-args e))
              (let ((lfun (cdr (assoc a env))))
                (when lfun (add-callgr-call fun lfun gr)))))
      (dolist (c (cexpr-conts e)) (fill-callgr c fun env gr)))))



;;;;
;;;; Finding Tail-Only Uses
;;;;

(defstruct (non-tail-table (:type list)
                           (:constructor make-non-tail-table (go-as-tail)))
  (hash (make-hash-table)) go-as-tail)

(defun non-tail-p (def tbl) (gethash def (non-tail-table-hash tbl)))

(defun mark-non-tail (var env tbl)
  (let ((def (cdr (assoc var env))))
    (when def (setf (gethash def (non-tail-table-hash tbl)) t))))

(defun compute-non-tail-table (e go-as-tail)
  (let ((tbl (make-non-tail-table go-as-tail)))
    (fill-non-tail-table e nil tbl)
    tbl))

(defun fill-non-tail-table (e env tbl)
  (cond
   ((letrec-p e)
    (let* ((args (letrec-args e))
           (vars (letrec-vars e))
           (nenv (append (mapcar #'cons vars args) env)))
      (dolist (a (letrec-args e)) (fill-non-tail-table a nenv tbl))
      (fill-non-tail-table (letrec-body e) nenv tbl)))
   ((call-p e)
    (let ((form (call-form e)))
      (dolist (a (call-args e)) (mark-non-tail a env tbl))
      (when form
            (mark-non-tail (call-fun e) env tbl)
            (fill-non-tail-table form env tbl))))
   (t (unless (and (non-tail-table-go-as-tail tbl) (primop-eq e 'go))
              (dolist (a (cexpr-args e)) (mark-non-tail a env tbl))
              (dolist (c (cexpr-conts e)) (fill-non-tail-table c env tbl))))))

(defun safe-use-p (fun root uses gr loc-ok)
  (let ((done (list root))
        (pending (list fun)))
    (loop
     (if (null pending)
         (return t)
       (let ((next (pop pending)))
         (unless (member next done)
                 (when (and (non-tail-p next uses)
                            (or (not loc-ok) (eq (func-type next) 'lambda)))
                       (return nil))
                 (push next done)
                 (setf pending (union (callgr-callers next gr) pending))))))))


;;;;
;;;; Removing Safe Blocks and Tagbodies
;;;;

(defun drop-safe-tags-one (e)
  (let ((gr (compute-callgr e t))
        (uses (compute-non-tail-table e t))
        (tags (compute-tags-table e)))
    (if (tags-table-has-tags tags)
        (drop-unused-tags (drop-safe-jumps e e uses tags gr) tags)
      e)))

(defun drop-safe-tags (e)
  (loop
   (let ((ne (drop-safe-tags-one e)))
     (if (eq e ne)
         (return e)
       (setf e ne)))))

(defstruct (tags-table-entry
            (:type list)
            (:constructor make-tags-table-entry (tag fun def)))
  tag fun def used)

(defun make-tags-table () (make-hash-table))

(defun lookup-tags-table-tag (tag table) (gethash tag table))

(defun lookup-tags-table-fun (fun table) (gethash fun table))

(defun add-tags-table-entry (fun env table)
  (let* ((def (cdr (assoc fun env)))
         (tag (first (func-vars def)))
         (entry (make-tags-table-entry tag fun def)))
    (setf (gethash tag table) entry)
    (setf (gethash fun table) entry)))

(defun tags-table-has-tags (table) (< 0 (hash-table-count table)))

(defun compute-tags-table (e)
  (let ((table (make-tags-table)))
    (fill-tags-table e nil table)
    table))

(defun fill-tags-table (e env table)
  (cond
   ((primop-p e)
    (let ((form (primop-form e)))
      (when (or (primop-eq e 'block) (primop-eq e 'tagbody))
            (add-tags-table-entry (first (primop-args e)) env table))
      (when form (fill-tags-table (primop-form e) env table))))
   ((letrec-p e)
    (let* ((args (letrec-args e))
           (vars (letrec-vars e))
           (nenv (append (mapcar #'cons vars args) env)))
      (dolist (a args) (fill-tags-table a nenv table))
      (fill-tags-table (letrec-body e) nenv table)))
   (t (dolist (c (cexpr-conts e)) (fill-tags-table c env table)))))

(defun drop-safe-jumps (e fun uses tags gr)
  (cond
   ((primop-p e) (drop-safe-jumps-primop e fun uses tags gr))
   ((letrec-p e) (drop-safe-jumps-letrec e fun uses tags gr))
   ((call-p e) (drop-safe-jumps-call e fun uses tags gr))
   (t (drop-safe-jumps-cexpr e fun uses tags gr))))

(defun drop-safe-jumps-primop (e fun uses tags gr)
  (let ((args (primop-args e)))
    (cond
     ((primop-eq e 'go)
      (let* ((tag (first args))
             (entry (lookup-tags-table-tag tag tags))
             (def (tags-table-entry-def entry))
             (ctx (primop-context e)))
        (cond
         ((safe-use-p fun def uses gr nil)
          (make-call (second args) nil nil nil nil nil ctx))
         (t (setf (tags-table-entry-used entry) t)
            e))))
     ((safe-return-from-cont-p e fun uses tags gr)
      (remake-cexpr e (cexpr-fun e) (cexpr-args e) nil nil))
     ((primop-eq e 'return-from)
      (let* ((tag (first args))
             (entry (lookup-tags-table-tag tag tags)))
        (setf (tags-table-entry-used entry) t)
        e))
     (t (drop-safe-jumps-cexpr e fun uses tags gr)))))

(defun safe-return-from-cont-p (e fun uses tags gr)
  (let ((form (first (cexpr-conts e))))
    (if (and form (primop-eq form 'return-from))
        (let* ((tag (first (primop-args form)))
               (entry (lookup-tags-table-tag tag tags))
               (def (tags-table-entry-def entry)))
          (safe-use-p fun def uses gr nil))
      nil)))

(defun drop-safe-jumps-letrec (e fun uses tags gr)
  (let ((nargs (mapcar #'(lambda (x) (drop-safe-jumps x x uses tags gr))
                       (letrec-args e)))
        (nbody (drop-safe-jumps (letrec-body e) fun uses tags gr)))
    (remake-cexpr e nil nargs (letrec-vars e) (list nbody))))

(defun drop-safe-jumps-cexpr (e fun uses tags gr)
  (let ((nconts (mapcar #'(lambda (x) (drop-safe-jumps x fun uses tags gr))
                        (cexpr-conts e))))
    (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts)))

(defun drop-safe-jumps-call (e fun uses tags gr)
  (if (safe-return-from-cont-p e fun uses tags gr)
      (remake-cexpr e (cexpr-fun e) (cexpr-args e) nil nil)
    (drop-safe-jumps-cexpr e fun uses tags gr)))

(defun drop-unused-tags (e tags)
  (cond
   ((letrec-p e)
    (let ((nargs (mapcar #'(lambda (x) (drop-unused-tags x tags))
                         (letrec-args e)))
          (nbody (drop-unused-tags (letrec-body e) tags)))
      (remake-cexpr e nil nargs (letrec-vars e) (list nbody))))
   ((or (primop-eq e 'block) (primop-eq e 'tagbody))
    (let* ((args (primop-args e))
           (vars (primop-vars e))
           (form (primop-form e))
           (nform (if form (drop-unused-tags form tags) nil))
           (nconts (if nform (list nform) nil))
           (fun (first args))
           (entry (lookup-tags-table-fun fun tags)))
      (if (tags-table-entry-used entry)
          (remake-cexpr e (primop-fun e) args vars nconts)
        (let ((ctx (primop-context e)))
          (make-call fun (list (make-const nil)) vars nform nil nil ctx)))))
   (t (let ((nconts (mapcar #'(lambda (x) (drop-unused-tags x tags))
                            (cexpr-conts e))))
        (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts)))))


;;;;
;;;; Removing Safe Cells
;;;;

(defstruct (cell-table-entry (:type list)
                             (:constructor make-cell-table-entry (cell fun)))
  cell fun unsafe)

(defstruct (cell-table (:type list))
  entries (hash (make-hash-table)))

(defun lookup-cell-table-entry (cell table)
  (gethash cell (cell-table-hash table)))

(defun add-cell-table-entry (cell fun table)
  (let ((entry (make-cell-table-entry cell fun)))
    (setf (gethash cell (cell-table-hash table)) entry)
    (push entry (cell-table-entries table))))  

(defun compute-cell-table (e uses gr)
  (let ((table (make-cell-table)))
    (fill-cell-table e e uses gr table)
    table))

(defun fill-cell-table (e fun uses gr table)
  (cond
   ((primop-p e)
    (case (primop-fun e)
          (make-cell (add-cell-table-entry (first (primop-vars e)) fun table))
          ((cell-value set-cell-value)
           (let* ((cell (first (primop-args e)))
                  (entry (lookup-cell-table-entry cell table))
                  (loc-ok (eq (primop-fun e) 'cell-value)))
             (unless (safe-use-p fun (cell-table-entry-fun entry) uses gr loc-ok)
                     (setf (cell-table-entry-unsafe entry) t)))))
    (let ((form (primop-form e)))
      (when form (fill-cell-table form fun uses gr table))))
   ((letrec-p e)
    (dolist (a (letrec-args e)) (fill-cell-table a a uses gr table))
    (fill-cell-table (letrec-body e) fun uses gr table))
   (t (dolist (c (cexpr-conts e)) (fill-cell-table c fun uses gr table)))))

(defun find-safe-cells (e gr)
  (let* ((uses (compute-non-tail-table e nil))
         (ctab (compute-cell-table e uses gr))
         (entries (cell-table-entries ctab))
         (safe-entries (remove-if #'cell-table-entry-unsafe entries)))
    (mapcar #'cell-table-entry-cell safe-entries)))

(defstruct (free-cell-entry (:type list)
                            (:constructor (make-free-cell-entry (def))))
  def free bound)

(defstruct (free-cell-table (:constructor make-free-cell-table (cells)))
  entries (hash (make-hash-table)) cells)

(defun add-free-cell-entry (entry table)
  (let ((def (free-cell-entry-def entry)))
    (setf (gethash def (free-cell-table-hash table)) entry)
    (push entry (free-cell-table-entries table))))

(defun lookup-free-cell-entry (def table)
  (gethash def (free-cell-table-hash table)))

(defun init-free-cell-table (e entry env table)
  (cond
   ((primop-p e)
    (case (primop-fun e)
          (make-cell
           (let ((cell (first (primop-vars e))))
             (when (member cell (free-cell-table-cells table))
                   (push cell (free-cell-entry-bound entry)))))
          ((cell-value set-cell-value)
           (let ((bound (free-cell-entry-bound entry))
                 (cell (first (cexpr-args e))))
             (when (and (member cell (free-cell-table-cells table))
                        (not (member cell bound)))
                   (pushnew cell (free-cell-entry-free entry))))))
    (let ((form (primop-form e)))
      (when form (init-free-cell-table form entry env table))))
   ((letrec-p e)
    (let* ((vars (letrec-vars e))
           (args (letrec-args e))
           (nenv (append (mapcar #'cons vars args) env))
           (entries (mapcar #'make-free-cell-entry args)))
      (dolist (e entries) (add-free-cell-entry e table))
      (do ((args args (rest args))
           (entries entries (rest entries)))
          ((or (null args) (null entries)))
          (init-free-cell-table (first args) (first entries) nenv table))
      (init-free-cell-table (letrec-body e) entry nenv table)))
   (t (dolist (c (cexpr-conts e)) (init-free-cell-table c entry env table)))))

(defun compute-free-cell-table (e)
  (let* ((gr (compute-callgr e))
         (cells (find-safe-cells e gr))
         (table (make-free-cell-table cells))
         (entry (make-free-cell-entry e)))
    (add-free-cell-entry entry table)
    (init-free-cell-table e entry nil table)
    (let ((funs (callgr-functions gr)))
      (loop
       (let ((changed nil))
         (dolist (fun funs)
           (let ((entry (lookup-free-cell-entry fun table)))
             (let ((free (free-cell-entry-free entry))
                   (bound (free-cell-entry-bound entry)))
               (dolist (c (callgr-callees (free-cell-entry-def entry) gr))
                 (dolist (v (free-cell-entry-free (lookup-free-cell-entry c table)))
                   (unless (or (member v free) (member v bound))
                           (push v free)
                           (setf changed t))))
               (setf (free-cell-entry-free entry) free))))
         (unless changed (return)))))
    table))

(defun drop-safe-cells (e)
  (let* ((ne (split-local-functions e))
         (table (compute-free-cell-table ne)))
    (if (free-cell-table-cells table)
        (drop-cells ne nil nil table)
      ne)))

(defun drop-cells (e env fenv table)
  (cond
   ((letrec-p e) (drop-cells-letrec e env fenv table))
   ((call-p e) (drop-cells-call e env fenv table))
   ((func-p e) (drop-cells-func e env fenv table))
   ((primop-p e)
    (case (primop-fun e)
          (make-cell (drop-cells-make-cell e env fenv table))
          (cell-value (drop-cells-cell-value e env fenv table))
          (set-cell-value (drop-cells-set-cell-value e env fenv table))
          (t (drop-cells-cexpr e env fenv table))))
   (t (drop-cells-cexpr e env fenv table))))

(defun drop-cells-cexpr (e env fenv table)
  (let* ((conts (cexpr-conts e))
         (nconts (mapcar #'(lambda (x) (drop-cells x env fenv table)) conts))
         (nargs (sublis env (cexpr-args e))))
    (remake-cexpr e (cexpr-fun e) nargs (cexpr-vars e) nconts)))

(defun drop-cells-letrec (e env fenv table)
  (let* ((vars (letrec-vars e))
         (args (letrec-args e))
         (nfenv (append (mapcar #'cons vars args) fenv))
         (nargs (mapcar #'(lambda (x) (drop-cells x env nfenv table)) args))
         (nbody (drop-cells (letrec-body e) env nfenv table)))
    (remake-cexpr e nil nargs vars (list nbody))))

(defun drop-cells-call (e env fenv table)
  (let* ((fun (call-fun e))
         (args (call-args e))
         (vars (call-vars e))
         (form (call-form e))
         (nconts (if form (list (drop-cells form env fenv table)) nil))
         (entry (lookup-free-cell-entry (cdr (assoc fun fenv)) table)))
    (if entry
        (let* ((free (free-cell-entry-free entry))
               (nargs (append (sublis env free) (sublis env args))))
          (remake-cexpr e fun nargs vars nconts))
      (remake-cexpr e (sublis env fun) (sublis env args) vars nconts))))

(defun drop-cells-func (e env fenv table)
  (let* ((entry (lookup-free-cell-entry e table))
         (free (free-cell-entry-free entry))
         (fvars (mapcar #'copy-var free))
         (nenv (append (mapcar #'cons free fvars) env))
         (nform (drop-cells (func-body e) nenv fenv table))
         (nvars (append fvars (func-vars e)))
         (ne (remake-cexpr e nil nil nvars (list nform))))
    (when fvars (incf (func-nreq ne) (length fvars)))
    ne))

(defun drop-cells-make-cell (e env fenv table)
  (let ((cell (first (primop-vars e))))
    (if (member cell (free-cell-table-cells table))
        (let* ((val (sublis env (first (primop-args e))))
               (nenv (cons (cons cell val) env)))
          (drop-cells (primop-form e) nenv fenv table))
      (drop-cells-cexpr e env fenv table))))

(defun drop-cells-cell-value (e env fenv table)
  (let ((cell (first (primop-args e))))
    (if (member cell (free-cell-table-cells table))
        (let ((var (first (primop-vars e)))
              (val (sublis env (cdr (assoc cell env))))
              (form (primop-form e)))
          (cond
           (var (let ((nenv (cons (cons var (sublis env val)) env)))
                  (drop-cells form nenv fenv table)))
           (form (drop-cells form env fenv table))
           (t (let ((ctx (primop-context e)))
                (make-primop 'values (list val) nil nil nil ctx)))))
      (drop-cells-cexpr e env fenv table))))

(defun drop-cells-set-cell-value (e env fenv table)
  (let* ((args (primop-args e))
         (cell (first args)))
    (if (member cell (free-cell-table-cells table))
        (let* ((val (sublis env (second args)))
               (nenv (cons (cons cell val) env)))
          (drop-cells (primop-form e) nenv fenv table))
      (drop-cells-cexpr e env fenv table))))
