(in-package "NEWCMP")

;;;;
;;;; Utility Functions
;;;;

(defun proper-list-p (list)
  (do ((list list (rest list)))
      ((null list) t)
      (unless (consp list) (return nil))))

(defun lambda-expression-p (x)
  (and (consp x)
       (eq (first x) 'lambda)
       (proper-list-p (second x))
       (proper-list-p (rest (rest x)))))

(defun split-decls-body (body doc)
  (let ((decls nil)
        (docstr (if doc nil "")))
    (dolist (e body)
      (cond
       ((stringp e) (if docstr (return) (setf docstr e)))
       ((and (consp e) (eq (first e) 'declare))
        (setf decls (append (rest e) decls)))
       (t (return)))
      (pop body))
    (values decls
            (if (null (rest body))
                (first body)
              (cons 'progn body))
            docstr)))

(defun not-inline-p (sym env)
  (eq (second (assoc 'inline (nth-value 2 (function-information sym env))))
      'notinline))

(defun inline-p (sym env)
  (eq (second (assoc 'inline (nth-value 2 (function-information sym env))))
      'inline))

(defun make-tmp-var (&optional (name "TMP") (type t))
  (let ((sym (if (and (symbolp name) (not (null name)))
                 name
               (gensym (if (stringp name) name "TMP")))))
    (make-var sym type nil)))

(defun make-var-list (syms ctx)
  (mapcar #'(lambda (s) (make-var s t ctx)) syms))
(defun simple-constant-p (e)
  (flet ((simple-value-p (v) (or (symbolp v) (numberp v) (characterp v))))
    (and (constantp e)
         (if (symbolp e)
             (simple-value-p (symbol-value e))
           (simple-value-p e)))))


;;;;
;;;; Macro Expansion
;;;;

(defun expand-expression-1 (e env)
  (if (and (consp e) (symbolp (first e)))
      (if (or (member (first e)
                      '(progv not null row-major-aref xlisp::%set-rm-aref))
              (nth-value 1 (function-information (first e) env))
              (not-inline-p (first e) env))
          (macroexpand-1 e env)
        (let ((entry (assoc (first e) xlisp::*cmp-global-macros*)))
          (if entry
              (let ((ne (macroexpand-1 e (list nil (list entry)))))
                ;;**** a hack for compiler macros -- need to fix ctx
                (register-global-call (first e) (rest e) nil)
                (if (eq e ne)
                    (macroexpand-1 e env)
                  (values ne t)))
            (macroexpand-1 e env))))
    (macroexpand-1 e env)))

(defun expand-expression (form env)
  (multiple-value-bind (form expanded)
                       (expand-expression-1 form env)
    (if expanded
        (loop
         (multiple-value-bind (newform expanded)
                              (expand-expression-1 form env)
           (unless expanded (return (values newform t)))
           (setf form newform)))
      (values form nil))))


;;;;
;;;; Compilation Context 
;;;;

(defun make-context (e) (declare (ignore e)) nil)
(defun augment-context (e ctx) (cons e ctx))


;;;;
;;;; Converter Environment
;;;;

(declaim (inline info-dict info-env make-info))
(defun info-dict (x) (car x))
(defun info-env (x) (cdr x))
(defun make-info (x y) (cons x y))

(defun add-dict-entries (names values type dict)
  (append (mapcar #'(lambda (x y) (list x type y)) names values) dict))

(defun find-dict-entry (name type dict)
  (dolist (d dict)
    (when (and (eql name (first d)) (eq type (second d)))
          (return (values (third d) t)))))

(defun add-local-vars (names nodes decls info)
  (let* ((vars (remove-if #'(lambda (x) (special-var-p x info)) names))
         (nenv (augment-environment (info-env info) :variable vars))
         (nnenv (augment-environment nenv :declare decls))
         (dict (add-dict-entries names nodes :variable (info-dict info))))
    (do ((names names (rest names))
         (nodes nodes (rest nodes)))
        ((or (null names) (null nodes)))
        (let* ((vdecs (nth-value 2 (variable-information (first names) nnenv)))
               (tentry (assoc 'type vdecs))
               (type (if tentry (second tentry) t))
               (ignore (eql (second (assoc 'ignore vdecs)) 'ignore)))
          (setf (var-type (first nodes)) (if ignore nil type))))
    (make-info dict nnenv)))           

(defun lookup-var (e info ctx)
  (let ((v (find-dict-entry e :variable (info-dict info))))
    (when (null (var-type v))
          (warn-ignored-variable-used v ctx)
          (setf (var-type v) t))
    (setf (var-tag v) t)
    v))

(defun local-var-p (e info)
  (and (symbolp e) (eq (variable-information e (info-env info)) :lexical)))

(defun add-local-funs (names nodes callinfos decls info)
  (declare (ignore decls))
  (let ((entries (mapcar #'list nodes callinfos)))
    (make-info (add-dict-entries names entries :function (info-dict info))
               (augment-environment (info-env info) :function names))))

(defun lookup-fun (e info)
  (let* ((entry (find-dict-entry e :function (info-dict info)))
         (v (first entry)))
    (setf (var-tag v) t)
    v))

(defun local-callinfo (e info)
  (let ((entry (find-dict-entry e :function (info-dict info))))
    (second entry)))

(defun local-fun-p (e info)
  (and (symbolp e) (nth-value 1 (function-information e (info-env info)))))

(defun add-block (name tag info)
  (make-info (add-dict-entries (list name) (list (list tag nil))
                               :block (info-dict info))
             (info-env info)))

(defun lookup-block (name info)
  (let ((val (find-dict-entry name :block (info-dict info))))
    (unless val (error "no block named ~s in current lexical context" name))
    (setf (second val) t)
    (first val)))

(defun block-used-p (name info)
  (second (find-dict-entry name :block (info-dict info))))

(defun add-go-tags (names tag funs info)
  (let ((tlist (mapcar #'(lambda (fun) (list tag fun)) funs)))
    (make-info (add-dict-entries names tlist :go (info-dict info))
               (info-env info))))

(defun lookup-go-tag (name info)
  (let ((val (find-dict-entry name :go (info-dict info))))
    (unless val (error "no tag named ~s in current lexical context" name))
    val))

(defun add-decls-info (decls info)
  (if decls
      (make-info (info-dict info)
                 (augment-environment (info-env info) :declare decls))
    info))

(defun special-var-p (e info)
  (and (symbolp e)
       (or (specialp e)
           (eq (variable-information e (info-env info)) :special))))


;;;;
;;;; ANF Converter
;;;;


;;;
;;; Main Driver Function
;;;

(defun convert (e info kv kf ctx)
  (let* ((actx (augment-context e ctx))
         (ee (expand-expression e (info-env info)))
         (ectx (augment-context ee actx)))
    (cond
     ((consp ee)
      (let ((fun (first ee))
            (args (rest ee)))
        (cond
         ((lambda-expression-p fun)
          (convert-lambda-call fun args info kv kf ectx))
         ((symbolp fun) (convert-symbol-call fun args info kv kf ectx))
         (t (error "~s is not a valid function" fun)))))
     ((symbolp ee) (convert-symbol ee info kv kf ectx))
     (t (convert-constant ee kv kf ectx)))))


;;;
;;; Converter Dispatch System
;;;

(let ((table (make-hash-table :test 'eq)))
  (defun add-converter (name fun) (setf (gethash name table) fun))
  (defun get-converter (name) (gethash name table)))

(defmacro define-converter (name args &body body)
  `(progn
     (add-converter ',name #'(lambda ,args ,@body))
     ',name))


;;;
;;; Converters for Compound Expressions
;;;

(defun convert-lambda-call (fun args info kv kf ctx)
  (let ((arg (convert-lambda nil fun info nil nil ctx))
        (var (make-tmp-var)))
    (check-lambda-argument-count fun args ctx)
    (let ((body (convert-call var args info kv kf ctx)))
      (make-letrec (list var) (list arg) body ctx))))

(defun convert-symbol-call (fun args info kv kf ctx)
  (cond
   ((local-fun-p fun info)
    (check-local-argument-count fun args info ctx)
    (convert-call (lookup-fun fun info) args info kv kf ctx))
   (t (let ((c (get-converter fun)))
        (cond
         (c (funcall c args info kv kf ctx))
         ((and (inline-p fun (info-env info)) (get-inline-definition fun))
          (convert-call-inline fun args info kv kf ctx))
         (t (check-global-argument-count fun args info ctx)
            (convert-call fun args info kv kf ctx)))))))

(defun convert-call (cf args info kv kf ctx)
  (let* ((argvars (mapcar #'make-tmp-var args))
         (vars (if (var-p kv) (list kv) nil))
         (vi (if (and kf (null kv)) t nil))
         (call (make-call cf argvars vars kf vi nil ctx)))
    (convert-call-args argvars args call info ctx)))

(defun convert-call-args (vars args call info ctx)
  (let ((form call)
        (rvars nil)
        (rargs nil)
        (env nil))
    (do ((vars vars (rest vars))
         (args args (rest args)))
        ((or (null vars) (null args)))
        (let ((ea (expand-expression (first args) (info-env info)))
              (v (first vars)))
          (cond
           ((and (symbolp ea) (local-var-p ea info))
            (push (cons v (lookup-var ea info ctx)) env))
           ((and (symbolp ea) (simple-constant-p ea))
            (push (cons v (make-const (symbol-value ea))) env))
           ((and (consp ea) (eq (first ea) 'quote) (null (nthcdr 3 ea)))
            (push (cons v (make-const (second ea))) env))
           ((and (not (symbolp ea)) (not (consp ea)))
            (push (cons v (make-const ea)) env))
           (t (push v rvars) (push ea rargs)))))
    (when env
          (setf form (copy-structure form))
          (setf (cexpr-args form) (sublis env (cexpr-args form))))
    (do ((rvars rvars (rest rvars))
         (rargs rargs (rest rargs)))
        ((or (null rvars) (null rargs)))
        (setf form (convert (first rargs) info (first rvars) form ctx)))
    form))



;;;
;;; Inlining Calls
;;;

(export '*inlining-enabled*)
(defvar *inlining-enabled* t)

(defvar *inlining* nil)

(defun convert-call-inline (fun args info kv kf ctx)
  (if *inlining-enabled*
      (let ((def (get-inline-definition fun)))
        (let ((val (check-call args (lambda-list-callinfo (second def)))))
          (when val (signal-argument-count-error fun val ctx)))
        (if (member fun *inlining*)
            (convert-call fun args info kv kf ctx)
          (let ((*inlining* (cons fun *inlining*)))
            (register-global-call fun args ctx)
            (let* ((form `(labels ((,fun ,@(rest def))) #',fun))
                   (cform (func-body (optimize (compile-front-form form nil))))
                   (fvars (letrec-vars cform))
                   (fargs (letrec-args cform))
                   (nbody (convert-call (first fvars) args info kv kf ctx)))
              (remake-cexpr cform nil fargs fvars (list nbody))))))
    (convert-call fun args info kv kf ctx)))



;;;
;;; Converters for Atoms
;;;

(defun convert-symbol (e info kv kf ctx)
  (cond
   ((local-var-p e info) (convert-value (lookup-var e info ctx) kv kf ctx))
   ((simple-constant-p e) (convert-constant (symbol-value e) kv kf ctx))
   (t (convert-special-variable e info kv kf ctx))))

(defun convert-constant (e kv kf ctx)
  (convert-value (make-const e) kv kf ctx))

(defun convert-value (v kv kf ctx)
  (cond
   ((or (null kf) (eql kv t)) (make-primop 'values (list v) nil kf nil ctx))
   ((null kv) kf)
   (t (make-bind (list kv) (list v) kf ctx))))

(defun convert-special-variable (e info kv kf ctx)
  (unless (special-var-p e info) (warn-undeclared-global e ctx))
  (convert-standard-primop 'symbol-value nil info kv kf e ctx))

(defun convert-standard-primop (which args info kv kf data ctx)
  (let* ((vars (mapcar #'make-tmp-var args))
         (vl (cond
              ((null kf) nil)
              ((var-p kv) (list kv))
              (t (list (make-tmp-var)))))
         (nkf (if (and kf (eq kv t))
                  (make-primop 'values vl nil kf nil ctx)
                kf))
         (form (make-primop which vars vl nkf data ctx)))
    (convert-call-args vars args form info ctx)))


;;;;
;;;; Lambda Expression Conversion
;;;;

(defun parse-lambda-list (lambda-list ctx)
  (unless (proper-list-p lambda-list) (bad-lambda-list lambda-list ctx))
  (flet ((chop (x y) (if y (ldiff x y) x)))
    (let* ((list0 lambda-list)
           (aux (member '&aux list0))
           (list1 (chop list0 aux))
           (other (member '&allow-other-keys list1))
           (list2 (chop list1 other))
           (key (member '&key list2))
           (list3 (chop list2 key))
           (rest (member '&rest list3))
           (list4 (chop list3 rest))
           (opt (member '&optional list4))
           (req (chop list4 opt)))
      (flet ((no-lkeys (x) (null (intersection x lambda-list-keywords))))
        (unless (and (no-lkeys (rest aux))
                     (null (rest other))
                     (or (null other) key)
                     (no-lkeys (rest key))
                     (or (null rest)
                         (and (= (length rest) 2) (symbolp (second rest))))
                     (no-lkeys (rest rest))
                     (no-lkeys (rest opt))
                     (no-lkeys req)
                     (every #'symbolp req))
                (bad-lambda-list lambda-list ctx)))
      (values req
              (parse-opt-args (rest opt) lambda-list ctx)
              (second rest)
              (if key t nil)
              (parse-key-args (rest key) lambda-list ctx)
              (if other t nil)
              (rest aux)))))

(defun bad-lambda-list (list ctx)
  (declare (ignore ctx))
  (error "~s is not a proper lambda list" list))

(defstruct (arginfo
            (:type list)
            (:constructor make-arginfo (sym val suppsym &optional key)))
  key sym val suppsym var suppvar)

(defun parse-opt-args (list lambda-list ctx)
  (let ((val nil))
    (dolist (a list (nreverse val))
      (cond
       ((symbolp a) (push (make-arginfo a nil nil) val))
       (t (unless (and (proper-list-p a)
                       (member (length a) '(1 2 3))
                       (symbolp (first a))
                       (symbolp (third a)))
                  (bad-lambda-list lambda-list ctx))
          (push (make-arginfo (first a) (second a) (third a)) val))))))

(defun parse-key-args (list lambda-list ctx)
  (flet ((mk-key (sym) (intern (symbol-name sym) "KEYWORD")))
    (let ((val nil))
      (dolist (a list (nreverse val))
        (cond
         ((symbolp a) (push (make-arginfo a nil nil (mk-key a)) val))
         ((proper-list-p a)
          (let ((arg (first a)))
            (unless (and (member (length a) '(1 2 3))
                         (or (symbolp arg)
                             (and (proper-list-p arg)
                                  (= (length arg) 2)
                                  (symbolp (first arg))
                                  (symbolp (second arg))))
                         (symbolp (third a)))
                    (bad-lambda-list lambda-list ctx))
            (let ((sym (if (symbolp arg) arg (second arg)))
                  (key (if (symbolp arg) (mk-key arg) (first arg))))
              (push (make-arginfo sym (second a) (third a) key) val))))
         (t (bad-lambda-list lambda-list ctx)))))))

(defun convert-lambda (name e info block have-block ctx)
  (multiple-value-bind (req opt rest has-key key other aux)
                       (parse-lambda-list (second e) ctx)
    (let ((lbody (multiple-value-bind (decls body)
                                      (split-decls-body (rest (rest e)) nil)
                   (let ((bbody (if have-block `(block ,block ,body) body)))
                     (if decls
                         (list (cons 'declare decls) bbody)
                       (list bbody))))))
      (convert-lambda-1 name req opt rest has-key key other aux lbody info
                        ctx))))

(defun convert-lambda-1 (name req opt rest has-key key other aux body info ctx)
  (let ((letform `(let* ,aux ,@body))
        (vars nil)
        (reqvars nil)
        (optvars nil)
        (restvar nil)
        (keylistvar nil))
    (when has-key
          (setf keylistvar (make-var (gensym) t ctx))
          (push keylistvar vars)
          (setf vars (proccess-opt-key-args letform key vars ctx)))
    (when (or rest has-key)
          (setf restvar (make-var (gensym) t ctx))
          (push restvar vars)
          (when rest (push `(,rest ,(var-name restvar)) (second letform))))
    (when opt (setf vars (proccess-opt-key-args letform opt vars ctx)))
    (dolist (v (reverse req))
      (let ((var (make-var (gensym) t ctx)))
        (push var reqvars)
        (push var vars)
        (push `(,v ,(var-name var)) (second letform))))
    (let* ((ninfo (add-local-vars (mapcar #'var-name vars) vars nil info))
           (cbody (convert letform ninfo nil nil ctx))
           (kargs (list keylistvar)))
      (when has-key
            (setf cbody
                  (make-primop 'check-last-keyarg kargs nil cbody other ctx))
            (dolist (k (reverse key))
              (let* ((vars (list (arginfo-var k) (arginfo-suppvar k)))
                     (val (arginfo-val k))
                     (data (list (arginfo-key k)
                                 (if (simple-constant-p val)
                                     (if (symbolp val) (symbol-value val) val)
                                   nil))))
                (setf cbody
                      (make-primop 'get-keyarg kargs vars cbody data ctx))))
            (setf cbody
                  (make-primop 'make-keyargs (list restvar) kargs cbody nil
                               ctx)))
      (dolist (k (reverse opt))
        (let* ((vars (list (arginfo-var k) (arginfo-suppvar k)))
               (val (arginfo-val k))
               (data (if (simple-constant-p val)
                         (if (symbolp val) (symbol-value val) val)
                       nil))
               (arg (make-var (gensym) t ctx)))
          (setf cbody (make-primop 'get-optarg (list arg) vars cbody data ctx))
          (push arg optvars)))
      (let ((list (if (or rest key) (list '&rest restvar) nil)))
        (when opt (setf list (append (cons '&optional optvars) list)))
        (setf list (append reqvars list))
        (make-func name list cbody 'lambda ctx)))))

(defun proccess-opt-key-args (letform args vars ctx)
  (dolist (k (reverse args) vars)
    (let ((var (make-var (gensym) t ctx))
          (suppvar (make-var (gensym) t ctx))
          (sym (arginfo-sym k))
          (suppsym (arginfo-suppsym k))
          (val (arginfo-val k)))
      (when suppsym (push `(,suppsym ,(var-name suppvar)) (second letform)))
      (if (simple-constant-p val)
          (push `(,sym ,(var-name var)) (second letform))
        (push `(,sym (if ,(var-name suppvar) ,(var-name var) ,val))
              (second letform)))
      (setf (arginfo-var k) var)
      (setf (arginfo-suppvar k) suppvar)
      (push var vars)
      (push suppvar vars))))


;;;;
;;;; Special Operator Converters
;;;;

(defconstant special-operators
  '(block      let*                  return-from      
    catch      load-time-value       setq             
    eval-when  locally               symbol-macrolet  
    flet       macrolet              tagbody          
    function   multiple-value-call   the              
    go         multiple-value-prog1  throw            
    if         progn                 unwind-protect   
    labels     progv                                
    let        quote))

(dolist (s (append '(case errset lambda nth-value) special-operators))
  (add-converter s
                 (let ((sym s))
                   #'(lambda (args info kv kf)
                       (declare (ignore args info kv kf))
                       (error "converter for ~a not implemented yet" sym)))))


;;
;; BLOCK Converter
;;

(define-converter block (args info kv kf ctx)
  (check-specop-argument-count 'block args 1 0 t)
  (let ((name (first args)))
    (check-block-name name 'block ctx)
    (let* ((tag (make-tmp-var))
           (bv (make-tmp-var))
           (ninfo (add-block name tag info))
           (body (convert (cons 'progn (rest args)) ninfo nil nil ctx))
           (vl (if (var-p kv) (list kv) nil))
           (vi (if (and kf (null kv)) t nil)))
      (if (block-used-p name ninfo)
          (let ((bfun (make-func nil (list tag) body 'lambda ctx))
                (nbody (make-primop 'block (list bv) vl kf nil ctx)))
            (make-letrec (list bv) (list bfun) nbody ctx))
        (let ((bfun (make-func nil nil body 'lambda ctx))
              (nbody (make-call bv nil vl kf vi nil ctx)))
          (make-letrec (list bv) (list bfun) nbody ctx))))))

(defun check-block-name (name operator ctx)
  (declare (ignore ctx))
  (unless (symbolp name)
          (error "~s is not a valid block name in ~s" name operator)))


;;
;; CASE Converter
;;

(define-converter case (args info kv kf ctx)
  (check-specop-argument-count 'case args 1 0 t)
  (let* ((clauses (case-add-default (rest args)))
         (conts (mapcar #'(lambda (x) (cons 'progn (rest x))) clauses))
         (data (coerce (butlast (mapcar #'first clauses)) 'vector)))
    (multiple-value-bind (k kfun)
                         (make-test-cont kv kf ctx)
      (let* ((v (make-tmp-var))
             (cconts (convert-test-conts conts info kv k ctx))
             (call (make-testop 'case (list v) cconts data ctx))
             (vl (list (first args)))
             (form (convert-call-args (list v) vl call info ctx)))
        (if k
            (make-letrec (list k) (list kfun) form ctx)
          form)))))

(defun case-add-default (clauses)
  (let ((val nil))
    (dolist (c clauses (nreverse (cons '(t nil) val)))
      (unless (and (proper-list-p c)
                   (or (not (consp (first c))) (proper-list-p (first c))))
              (error "bad case clause: ~s" c))
      (when (member (first c) '(t otherwise))
            (when (rest (member c clauses))
                  (issue-style-warning "OTHERWISE clause ~s is not last" c))
            (return (nreverse (cons c val))))
      (push c val))))

(defun make-test-cont (kv kf ctx)
  (if kf
      (values (make-tmp-var)
              (make-func nil (if (var-p kv) (list kv) nil) kf 'lambda ctx))
    (values nil nil)))

(defun convert-test-conts (conts info kv k ctx)
  (if k
      (mapcar #'(lambda (cont)
                  (let* ((ckv (if (var-p kv) (copy-var kv) kv))
                         (ckf (make-call k (if (var-p kv) (list ckv) nil)
                                         nil nil nil nil ctx)))
                    (convert cont info ckv ckf ctx)))
              conts)
    (mapcar #'(lambda (cont) (convert cont info nil nil ctx)) conts)))


;;
;; CATCH Converter
;;

(define-converter catch (args info kv kf ctx)
  (check-specop-argument-count 'catch args 1 0 t)
  (let* ((tag (first args))
         (tv (make-tmp-var))
         (bv (make-tmp-var))
         (body (convert (cons 'progn (rest args)) info nil nil ctx))
         (bfun (make-func nil nil body 'lambda ctx))
         (vl (if (var-p kv) (list kv) nil))
         (catch (make-primop 'catch (list tv bv) vl kf nil ctx))
         (tkf (make-letrec (list bv) (list bfun) catch ctx)))
    (convert tag info tv tkf ctx)))


;;
;; ERRSET Converter
;;

(define-converter errset (args info kv kf ctx)
  (check-specop-argument-count 'errset args 1 1 nil)
  (let* ((pflag (if (consp (rest args)) (second args) t))
         (pv (make-tmp-var))
         (bv (make-tmp-var))
         (bfbody (convert (first args) info nil nil ctx))
         (bfun (make-func nil nil bfbody 'lambda ctx))
         (vl (if (var-p kv) (list kv) nil))
         (body (make-primop 'errset (list bv pv) vl kf nil ctx))
         (nkf (make-letrec (list bv) (list bfun) body ctx)))
    (convert pflag info pv nkf ctx)))


;;
;; FLET Converter
;;

(define-converter flet (args info kv kf ctx)
  (convert-funs args info kv kf 'flet ctx))

(defun convert-funs (args info kv kf which ctx)
  (check-specop-argument-count which args 1 0 t)
  (let* ((funs (first args))
         (names (mapcar #'first funs))
         (cis (mapcar #'(lambda (x) (lambda-list-callinfo (second x))) funs)))
    (multiple-value-bind (decls body)
                         (split-decls-body (rest args) nil)
      (let* ((fvars (make-var-list names ctx))
             (ninfo (add-local-funs names fvars cis decls info))
             (nninfo (add-decls-info decls ninfo))
             (cinfo (if (eq which 'labels) ninfo info)))
        (flet ((cvtfun (f)
                 (let ((n (first f))
                       (e (cons 'lambda (rest f))))
                   (convert-lambda n e cinfo n t ctx))))
          (let ((fargs (mapcar #'cvtfun funs))
                (bform (convert body nninfo kv kf ctx)))
            (check-unused-funs fvars ctx)
            (make-letrec fvars fargs bform ctx)))))))


;;
;; FUNCTION Converter
;;

(define-converter function (args info kv kf ctx)
  (check-specop-argument-count 'function args 1 0 nil)
  (let ((fun (first args)))
    (cond
     ((local-fun-p fun info)
      (convert-value (lookup-fun fun info) kv kf ctx))
     ((symbolp fun)
      (convert-standard-primop 'symbol-function nil info kv kf fun ctx))
     ((lambda-expression-p fun)
      (let* ((v (make-tmp-var)) ;;**** type??
             (cf (convert-lambda nil fun info nil nil ctx)))
        (make-letrec (list v) (list cf) (convert-value v kv kf ctx) ctx)))
     (t (error "bad argument for FUNCTION - ~s" fun)))))



;;
;; GO Converter
;;

(define-converter go (args info kv kf ctx)
  (declare (ignore kv kf))
  (check-specop-argument-count 'go args 1 0 nil)
  (let ((name (first args)))
    (check-tagbody-tag name 'go ctx)
    (let ((vars (lookup-go-tag name info)))
      (make-primop 'go vars nil nil nil ctx))))

(defun check-tagbody-tag (tag operator ctx)
  (declare (ignore ctx))
  (unless (tagbody-tag-p tag) (error "~s is not a valid ~s tag" tag operator)))
(defun tagbody-tag-p (val)
  (or (symbolp val) (integerp val)))


;;
;; IF Converter
;;
;;

(define-converter if (args info kv kf ctx)
  (check-specop-argument-count 'if args 2 1 nil)
  (let ((test (first args))
        (consequent (second args))
        (alternative (third args)))
    (if (constantp test)
        (convert (if test consequent alternative) info kv kf ctx)
      (multiple-value-bind (k kfun)
                           (make-test-cont kv kf ctx)
        (let* ((v (make-tmp-var))
               (conts (list consequent alternative))
               (cconts (convert-test-conts conts info kv k ctx))
               (call (make-testop t (list v) cconts nil ctx))
               (form (convert-call-args (list v) (list test) call info ctx)))
          (if k
              (make-letrec (list k) (list kfun) form ctx)
            form))))))



;;
;; LABELS Converter
;;

(define-converter labels (args info kv kf ctx)
  (convert-funs args info kv kf 'labels ctx))



;;
;; LAMBDA Converter
;;

(define-converter lambda (args info kv kf ctx)
  (convert `(function (lambda ,@args)) info kv kf ctx))


;;
;; LET Converter
;;

(define-converter let (args info kv kf ctx)
  (check-specop-argument-count 'let args 1 0 t)
  (let ((binds (canonicalize-bindings 'let (first args) ctx)))
    (multiple-value-bind (decls body)
                         (split-decls-body (rest args) nil)
      (let* ((vars (mapcar #'first binds))
             (args (mapcar #'second binds))
             (nodes (make-var-list vars ctx))
             (ninfo (add-local-vars vars nodes decls info))
             (specs (find-special-variable-nodes nodes ninfo))
             (form (convert-let-body body specs ninfo kv kf ctx)))
        (check-unused-vars (set-difference nodes specs) ninfo ctx)
        (do ((rnodes (reverse nodes) (rest rnodes))
             (rargs (reverse args) (rest rargs)))
            ((or (null rnodes) (null rargs)))
            (setf form (convert (first rargs) info (first rnodes) form ctx)))
        form))))

(defun canonicalize-bindings (fun binds ctx)
  (declare (ignore ctx))
  (unless (proper-list-p binds) (error "not a proper bindings list: ~s" binds))
  (flet ((canon (b)
           (when (and (symbolp b) (not (eq fun 'symbol-macrolet)))
                 (setf b (list b nil)))
           (unless (and (proper-list-p b)
                        (symbolp (first b))
                        (null (rest (rest b))))
                   (error "~s is not a proper ~s binding form" b fun))
           (when (constantp (first b))
                 (error "can't bind to the constant ~s" (first b)))
           b))
    (mapcar #'canon binds)))

(defun find-special-variable-nodes (nodes info)
  (remove-if-not #'(lambda (x) (special-var-p (var-name x) info)) nodes))

(defun convert-let-body (body specs info kv kf ctx)
  (if specs
      (let* ((denv (make-tmp-var))
             (nkf (make-primop 'dynamic-unbind (list denv) nil kf nil ctx))
             (form (convert body info (if kf kv t) nkf ctx)))
        (dolist (s (reverse specs))
          (setf form (make-primop 'set-symbol-value (list s) nil form
                                  (var-name s) ctx)))
        (let ((syms (make-const (mapcar #'var-name specs)))
              (tc (make-const t)))
          (make-primop 'dynamic-bind (list syms tc) (list denv) form nil ctx)))
    (convert body info kv kf ctx)))


;;
;; LET* Converter
;;

(define-converter let* (args info kv kf ctx)
  (check-specop-argument-count 'let* args 1 0 t)
  (let ((binds (canonicalize-bindings 'let* (first args) ctx)))
    (multiple-value-bind (decls body)
                         (split-decls-body (rest args) nil)
      (let* ((vars (mapcar #'first binds))
             (args (mapcar #'second binds))
             (nodes (make-var-list vars ctx))
             (ninfo (add-local-vars vars nodes decls info))
             (ilist (make-let*-infos nodes info ninfo))
             (specs (find-special-variable-nodes nodes ninfo))
             (denvs (mapcar #'make-tmp-var specs))
             (dalist (mapcar #'cons specs denvs))
             (form (convert-let*-body body ninfo kv kf denvs ctx)))
        (mapc #'(lambda (v arg info)
                  (setf form (convert-let*-arg v arg (cdr (assoc v dalist))
                                               info form ctx)))
              (reverse nodes)
              (reverse args)
              (reverse ilist))
        (check-unused-vars (set-difference nodes specs) ninfo ctx)
        form))))

(defun make-let*-infos (nodes info ninfo)
  (let ((ilist (list info)))
    (dolist (v (butlast nodes) (reverse ilist))
      (let* ((vn (var-name v))
             (env (info-env ninfo))
             (vi (multiple-value-list (variable-information vn env)))
             (decls (if (eq (first vi) :special)
                        (cons `(special ,vn) (third vi))
                      (third vi)))
             (ninfo (add-local-vars (list vn) (list v) decls info)))
        (setf info ninfo)
        (push info ilist)))))

(defun convert-let*-body (body info kv kf denvs ctx)
  (if denvs
      (let ((form kf))
        (dolist (de denvs)
          (setf form (make-primop 'dynamic-unbind (list de) nil form nil ctx)))
        (convert body info (if kf kv t) form ctx))
    (convert body info kv kf ctx)))

(defun convert-let*-arg (v arg denv info form ctx)
  (if denv
      (let* ((name (var-name v))
             (sc (make-const (list name)))
             (tc (make-const t))
             (nkf (make-primop 'set-symbol-value (list v) nil form name ctx))
             (nnkf (make-primop 'dynamic-bind (list sc tc) (list denv) nkf
                                nil ctx)))
        (convert arg info v nnkf ctx))
    (convert arg info v form ctx)))


;;
;; LOAD-TIME-VALUE Converter
;;



;;
;; LOCALLY Converter
;;

(define-converter locally (args info kv kf ctx)
  (multiple-value-bind (decls body)
                       (split-decls-body args nil)
    (convert body (add-decls-info decls info) kv kf ctx)))


;;
;; MACROLET Converter
;;

(define-converter macrolet (args info kv kf ctx)
  (check-specop-argument-count 'macrolet args 1 0 t)
  (let ((env (info-env info)))
    (flet ((mkmacro (x)
             (let ((name (first x))
                   (args (second x))
                   (body (rest (rest x))))
               (list name (parse-macro name args body env)))))
      (multiple-value-bind (decls body)
                           (split-decls-body (rest args) nil)
        (let* ((macs (mapcar #'mkmacro (first args)))
               (dict (info-dict info))
               (ninfo (make-info dict (augment-environment env :macro macs)))
               (nninfo (add-decls-info decls ninfo)))
          (convert body nninfo kv kf ctx))))))


;;
;; MULTIPLE-VALUE-CALL
;;

(define-converter multiple-value-call (args info kv kf ctx)
  (check-specop-argument-count 'multiple-value-call args 1 0 t)
  (let ((fun (first args))
        (forms (rest args)))
    (cond
     ((null forms) (convert `(funcall ,fun) info kv kf ctx))
     ((null (rest forms))
      (let* ((fv (make-tmp-var)) ;;****type?
             (vars (if (var-p kv) (list kv) nil))
             (vi (if (and kf (null kv)) t nil))
             (call (make-call fv nil vars kf vi t ctx)))
        (convert fun info fv (convert (first forms) info t call ctx) ctx)))
     (t (let ((vforms (mapcar #'(lambda (x) `(multiple-value-list ,x)) forms)))
          (convert `(apply ,fun (append ,@vforms)) info kv kf ctx))))))



;;
;; MULTIPLE-VALUE-PROG1
;;

(define-converter multiple-value-prog1 (args info kv kf ctx)
  (check-specop-argument-count 'multiple-value-prog1 args 1 0 t)
  (let ((first (first args))
        (rest (rest args)))
    (cond
     ((null rest) (convert first info kv kf ctx))
     ((and kf (or (var-p kv) (null kv)))
      (convert `(prog1 ,@args) info kv kf ctx))
     (t (let* ((vars (make-save-values-args))
               (nkf (make-primop 'restore-values vars nil kf nil ctx))
               (restform (convert (cons 'progn rest) info nil nkf ctx))
               (nnkf (make-primop 'save-values nil vars restform nil ctx)))
          (convert first info t nnkf ctx))))))

(defun make-save-values-args ()
  (list (make-tmp-var "VCOUNT") (make-tmp-var "VAL") (make-tmp-var "VREST")))


;;
;; NTH-VALUE Converter
;;

(define-converter nth-value (args info kv kf ctx)
  (check-specop-argument-count 'nth-value args 2 0 nil)
  (if (and (not (null kf)) (null kv))
      (let ((nkf (convert (second args) info nil kf ctx)))
        (convert (first args) info nil nkf ctx))
    (let* ((nv (make-tmp-var))
           (vars (if kv (list kv) nil))
           (form (make-primop 'nth-value (list nv) vars kf nil ctx))
           (nkf (convert (second args) info t form ctx)))
      (convert (first args) info nv nkf ctx))))


;;
;; PROGN Converter
;;

(define-converter progn (args info kv kf ctx)
  (check-specop-argument-count 'progn args 0 0 t)
  (let* ((rargs (reverse args))
         (val (convert (first rargs) info kv kf ctx)))
    (dolist (a (rest rargs) val)
      (setf val (convert a info nil val ctx)))))


;;
;; PROGV Converter
;;

(define-converter progv (args info kv kf ctx)
  (check-specop-argument-count 'progv args 2 0 t)
  (let* ((v (make-tmp-var))
         (bargs (list (first args) (second args)))
         (body (cons 'progn (rest (rest args))))
         (nkv (if (null kf) t kv))
         (nkf (make-primop 'dynamic-unbind (list v) nil kf nil ctx))
         (form (convert body info nkv nkf ctx)))
    (convert-standard-primop 'dynamic-bind bargs info v form nil ctx)))


;;
;; QUOTE Converter
;;

(define-converter quote (args info kv kf ctx)
  (declare (ignore info))
  (check-specop-argument-count 'quote args 1 0 nil)
  (convert-constant (first args) kv kf ctx))



;;
;; RETURN-FROM Converter
;;

(define-converter return-from (args info kv kf ctx)
  (declare (ignore kv kf))
  (check-specop-argument-count 'return-from args 1 1 nil)
  (let ((name (first args)))
    (check-block-name name 'return-from ctx)
    (let* ((tag (lookup-block name info))
           (nkf (make-primop 'return-from (list tag) nil nil nil ctx)))
      (convert (second args) info t nkf ctx))))


;;
;; SETQ Converter
;;

(define-converter setq (args info kv kf ctx)
  (check-specop-argument-count 'setq args 2 0 t)
  (let ((sym (first args))
        (val (second args))
        (rest (rest (rest args))))
    (unless (symbolp sym) (error "argument ~s to SETQ is not a symbol" sym))
    (cond
     ((rest (rest args))
      (convert `(progn (setq ,sym ,val) (setq ,@rest)) info kv kf ctx))
     ((eq (variable-information sym (info-env info)) :symbol-macro)
      (convert `(setf ,sym ,val) info kv kf ctx))
     (t (let* ((v (make-tmp-var))
               (nkf (convert-value v kv kf ctx)))
          (if (local-var-p sym info)
              (let ((var (lookup-var sym info ctx)))
                (convert val info v
                         (make-primop 'setq (list var v) nil nkf nil ctx)
                         ctx))
            (progn
              (unless (special-var-p sym info)
                      (warn-undeclared-global sym ctx))
              (convert val info v
                       (make-primop 'set-symbol-value (list v) nil nkf sym ctx)
                       ctx))))))))


;;
;; SYMBOL-MACROLET Converter
;;

(define-converter symbol-macrolet (args info kv kf ctx)
  (check-specop-argument-count 'symbol-macrolet args 1 0 t)
  (let ((binds (canonicalize-bindings 'symbol-macrolet (first args) ctx)))
    (multiple-value-bind (decls body)
                         (split-decls-body (rest args) nil)
      (let* ((nenv (augment-environment (info-env info) :symbol-macro binds))
             (ninfo (make-info (info-dict info) nenv))
             (nninfo (add-decls-info decls ninfo)))
        (convert body nninfo kv kf ctx)))))



;;
;; TAGBODY Converter
;;

(define-converter tagbody (args info kv kf ctx)
  (dolist (a args) (unless (consp a) (check-tagbody-tag a 'tagbody ctx)))
  (if (find-if #'tagbody-tag-p args)
      (convert-tagbody args info kv kf ctx)
    (convert `(progn ,@args nil) info kv kf ctx)))

(defun convert-tagbody (body info kv kf ctx)
  (let* ((split (split-tagbody body))
         (tag (make-tmp-var))
         (tags (reverse (mapcar #'first split)))
         (vars (make-var-list tags ctx))
         (ninfo (add-go-tags tags tag vars info))
         (svar (make-tmp-var))
         (args (make-tagbody-funs (mapcar #'rest split) (reverse vars) ninfo
                                  ctx))
         (sbody (make-letrec vars args
                             (make-call (first vars) nil nil nil nil nil ctx)
                             ctx))
         (sfun (make-func nil (list tag) sbody 'lambda ctx))
         (vl (if (var-p kv) (list kv) nil))
         (body (make-primop 'tagbody (list svar) vl kf nil ctx)))
    (make-letrec (list svar) (list sfun) body ctx)))

(defun split-tagbody (body)
  (unless (tagbody-tag-p (first body)) (push (gensym) body))
  (let ((val nil))
    (loop
     (let* ((next (member-if #'tagbody-tag-p (rest body))))
       (push (ldiff body next) val)
       (unless next (return))
       (setf body next)))
    val))

(defun make-tagbody-funs (bodies vars info ctx)
  (let* ((kf (convert-constant nil nil nil ctx))
         (body (convert `(progn ,@(first bodies)) info nil kf ctx))
         (name (var-name (first vars)))
         (iname (if (symbol-package name) name nil))
         (args (list (make-func iname nil body 'lambda ctx))))
    (do ((bodies (rest bodies) (rest bodies))
         (vars vars (rest vars)))
        ((null bodies))
        (let* ((kf (make-call (first vars) nil nil nil nil nil ctx))
               (name (var-name (second vars)))
               (body (convert `(progn ,@(first bodies)) info nil kf ctx)))
          (push (make-func name nil body 'lambda ctx) args)))
    args))


;;
;; THE Converter
;;

(define-converter the (args info kv kf ctx)
  (check-specop-argument-count 'the args 2 0 nil)
  (let ((type (first args))
        (form (second args)))
    (check-type-spec type info)
    (convert form info kv kf ctx)))


;;
;; THROW Converter
;;

(define-converter throw (args info kv kf ctx)
  (declare (ignore kv kf))
  (check-specop-argument-count 'throw args 2 0 nil)
  (let* ((tv (make-tmp-var))
         (tform (make-primop 'throw (list tv) nil nil nil ctx))
         (nkf (convert (second args) info t tform ctx)))
    (convert (first args) info tv nkf ctx)))


;;
;; UNWIND-PROTECT Converter
;;

(define-converter unwind-protect (args info kv kf ctx)
  (check-specop-argument-count 'unwind-protect args 1 0 t)
  (let ((form (first args)))
    (if (rest args)
        (let* ((bv (make-tmp-var))
               (pv (make-tmp-var))
               (pform (cons 'progn (rest args)))
               (cform (convert form info nil nil ctx))
               (cpform (convert pform info nil nil ctx))
               (bfun (make-func nil nil cform 'lambda ctx))
               (pfun (make-func nil nil cpform 'lambda ctx))
               (vl (if (var-p kv) (list kv) nil))
               (body (make-primop 'unwind-protect (list bv pv) vl kf nil ctx)))
          (make-letrec (list bv pv) (list bfun pfun) body ctx))
      (convert form info kv kf ctx))))


;;;;
;;;; Error Checking
;;;;

(defun issue-style-warning (string &rest args)
  (warn (make-condition 'style-warning
                        :format-string string
                        :format-arguments args)))

(export '*warn-undeclared-globals*)
(defvar *warn-undeclared-globals* nil)

(defun warn-undeclared-global (sym ctx)
  (declare (ignore ctx))
  (when (and *warn-undeclared-globals*
             #+xlisp-stat (not (member sym user::*variables*)))
        (issue-style-warning
         "the global variable ~s has not been declared special" sym)))

(defun check-unused-vars (vars info ctx)
  (let ((val nil))
    (dolist (v vars)
      (unless (or (var-tag v) (var-ignorable v info)) (push (var-name v) val)))
    (when val (warn-unused-variables val ctx))))

(defun var-ignorable (v info)
  (or (null (var-type v))
      (let* ((name (var-name v))
             (decls (nth-value 2 (variable-information name (info-env info)))))
        (member (second (assoc 'ignore decls)) '(ignore ignorable)))))

(defun warn-unused-variables (vars ctx)
  (declare (ignore ctx))
  (let ((v (first vars))
        (rv (rest vars)))
    (issue-style-warning "unused local variable~:[~;s~] ~s~{, ~s~}" rv v rv)))

(defun check-unused-funs (vars ctx)
  (let ((val nil))
    (dolist (v vars) (unless (var-tag v) (push (var-name v) val)))
    (when val (warn-unused-functions val ctx))))

(defun warn-unused-functions (funs ctx)
  (declare (ignore ctx))
  (let ((v (first funs))
        (rv (rest funs)))
    (issue-style-warning "unused local function~:[~;s~] ~s~{, ~s~}" rv v rv)))

(defun warn-ignored-variable-used (v ctx)
  (declare (ignore ctx))
  (issue-style-warning "attempt to access variable ~a declared IGNORE"
                       (var-name v)))

(defun check-specop-argument-count (op args nreq nopt restp)
  (let ((n (length args))
        (nro (+ nreq nopt)))
    (unless (<= nreq n) (error "too few arguments for ~s" op))
    (unless (or restp (<= n nro)) (error "too many arguments for ~s" op))))

(defun too-few-arguments (e m)
  (declare (ignore m))
  (error "too few arguments for ~s" (var-name (call-fun e))))

(defun too-many-arguments (e m)
  (declare (ignore m))
  (error "too many arguments for ~s" (var-name (call-fun e))))

(defun check-type-spec (type info)
  (declare (ignore type info))
  ;;**** do something useful
  t)

(defstruct (callinfo
            (:type list)
            (:constructor make-callinfo (nreq nopt restp keyp keys other)))
  nreq nopt restp keyp keys other)

(defun lambda-list-callinfo (list)
  (multiple-value-bind (req opt rest has-key key other)
                       (parse-lambda-list list nil)
    (make-callinfo (length req) (length opt) (if rest t nil)
                   has-key (mapcar #'arginfo-key key) other)))

(defun check-call (args ci &optional (check-keys *strict-keywords*))
  (let* ((nreq (callinfo-nreq ci))
         (nro (+ nreq (callinfo-nopt ci)))
         (restp (callinfo-restp ci))
         (keyp (callinfo-keyp ci))
         (n (if (listp args) (length args) args))) ;;**** allow for just argc
    (cond
     ((< n nreq) :too-few)
     ((and (< nro n) (not restp) (not keyp)) :too-many)
     ((and keyp (oddp (- n nro))) :missing-keyval)
     ((and keyp
           (not (callinfo-other ci))
           check-keys
           (let ((keyargs (nthcdr (if restp (1+ nro) nro) args))
                 (keys (callinfo-keys ci))
                 (badkey nil))
             (do ((args keyargs (cddr args)))
                 ((null args) badkey)
                 (let ((k (first args)))
                   (when (constantp k)
                         (cond
                          ((eq k :allow-other-keys) (return nil)) ;; maybe ok - stop check
                          ((not (member k keys)) (setf badkey t))))))))
      :invalid-keyword)
     (t nil))))

(defun check-lambda-argument-count (fun args ctx)
  (when (null (rest fun)) (error "bad lambda expression: ~s" fun))
  (let ((val (check-call args (lambda-list-callinfo (second fun)))))
    (when val (signal-argument-count-error fun val ctx))))

(defun signal-argument-count-error (fun which ctx)
  (declare (ignore ctx))
  (case which
        (:too-few (error "too few arguments for ~s" fun))
        (:too-many (error "too many arguments for ~s" fun))
        (:missing-keyval (error "missing keyword value for ~s" fun))
        (:invalid-keyword (error "invalid keyword in call to ~s" fun))
        (t (error "bad call to ~s" fun))))

(defun check-local-argument-count (fun args info ctx)
  (let ((val (check-call args (local-callinfo fun info))))
    (when val (signal-argument-count-error fun val ctx))))

(defun check-global-argument-count (fun args info ctx)
  (declare (ignore info))
  (let ((ci (global-callinfo fun)))
    (cond
     (ci (register-global-call fun args ctx)
         (when (check-call args ci)
               (issue-style-warning "call to ~s does not match definition"
                                    fun)))
     (t (let ((ci (sysfun-callinfo fun)))
          (if ci
              (let* ((check-keys (or *strict-keywords*
                                     (and (fboundp fun)
                                          (typep (symbol-function fun)
                                                 'subr))))
                     (val (check-call args ci check-keys)))
                (when val (signal-argument-count-error fun val ctx)))
            (register-global-call fun args ctx)))))))


;;;;
;;;; Register Built-In Functions
;;;;

(defparameter *sysfun-callinfo* (make-hash-table))

(defun sysfun-callinfo (fun) (gethash fun *sysfun-callinfo*))

(defmacro define-sysfun-callinfo (name nreq &optional (nopt 0) restp keyp keys
                                       other)
  (if (consp name)
      `(let ((ci (make-callinfo ,nreq ,nopt ,restp ,keyp ',keys ,other)))
         (dolist (s ',name)
           (unless (fboundp s) (warn "built-in function ~s is not defined" s))
           (setf (gethash s *sysfun-callinfo*) ci)))
    `(progn
       (unless (fboundp ',name)
               (warn "built-in function ~s is not defined" ',name))
       (setf (gethash ',name *sysfun-callinfo*)
             (make-callinfo ,nreq ,nopt ,restp ,keyp ',keys ,other))
       ',name)))

;; **** get-universal-time
(define-sysfun-callinfo
  (get-decoded-time get-internal-real-time get-internal-run-time
 list-all-packages)
  0)

(define-sysfun-callinfo make-pathname
  0 0 nil t (:case :defaults :device :directory :host :name :type :version))

(define-sysfun-callinfo make-hash-table
  0 0 nil t (:rehash-size :rehash-threshold :size :test))

(define-sysfun-callinfo
  (+ * /= < <= = > >= append break char-equal char-greaterp char-lessp
   char-not-equal char-not-greaterp char-not-lessp char< char<= char= char>
   char>= gcd lcm list list* logand logeqv logior logxor max min nconc values
   vector y-or-n-p yes-or-no-p)
  0 0 t)

;;**** listen nbutlast
(define-sysfun-callinfo
  (abort compute-restarts dribble finish-output force-output fresh-line
   gensym make-string-output-stream muffle-warning
   terpri)
  0 1)

;;**** copy-readtable
(define-sysfun-callinfo ( gentemp make-random-state) 0 2)

;;**** read-char-no-hang
(define-sysfun-callinfo
  (read read-char read-from-string read-line
   read-preserving-whitespace)
  0 4)
 
(define-sysfun-callinfo peek-char 0 5)
;;**** character disassemble fdefinition file-author fill-pointer
;;**** float-digits float-precision float-radix formatter graphic-char-p 
;;**** host-namestring inspect integer-decode-float interactive-stream-p
;;**** name-char pathnamep rationalize readtable-case readtablep
;;**** simple-bit-vector-p simple-string-p simple-vector-p sleep
;;**** special-operator-p standard-char-p step stream-element-type
;;**** stream-external-format sxhash vector-pop
(define-sysfun-callinfo
  (1+ 1- abs acos acosh alpha-char-p alphanumericp array-dimensions
   array-element-type array-rank array-total-size arrayp asin asinh
   atanh atom both-case-p boundp caaar caadr caar cadar caddr cadr car
   cdaar cdadr cdar cddar cdddr cddr cdr char-code char-downcase char-int
   char-upcase characterp cis clrhash code-char compiled-function-p complement
   complexp conjugate consp constantp copy-alist copy-list copy-seq
   copy-structure cos cosh xlisp::coerce-to-macro delete-file delete-package
   denominator directory-namestring eighth endp eval evenp exp fboundp fifth
   file-length file-namestring file-write-date find-all-symbols find-package
   first floatp fmakunbound fourth function-lambda-expression functionp
   get-output-stream-string hash-table-count hash-table-p
   hash-table-rehash-size hash-table-rehash-threshold hash-table-size
   hash-table-test identity imagpart input-stream-p integer-length integerp
   invoke-debugger invoke-restart-interactively isqrt keywordp length
   list-length listp logcount lognot lower-case-p make-symbol makunbound
   minusp namestring ninth not nreverse null numberp numerator oddp
   open-stream-p output-stream-p package-name package-nicknames
   package-shadowing-symbols package-use-list package-used-by-list
   packagep phase plusp prin1-to-string princ-to-string probe-file proclaim
   provide random-state-p rational rationalp realp realpart rest restart-name
   reverse second seventh signum sin sinh sixth sqrt streamp string stringp
   symbol-function symbol-name symbol-package symbol-plist symbol-value
   symbolp tan tanh tenth third truename type-of upper-case-p values-list 
   vectorp zerop xlisp::specialp xlisp::%copy-struct
   xlisp::byte-code-close)
  1)

(define-sysfun-callinfo (funcall) 1 0 t)

(define-sysfun-callinfo augment-environment
  1 0 nil t (:declare :function :macro :symbol-macro :variable))

(define-sysfun-callinfo close 1 0 nil t (:abort))

(define-sysfun-callinfo (delete-duplicates remove-duplicates)
  1 0 nil t (:from-end :test :test-not :start :end :key))

(define-sysfun-callinfo directory 1 0 nil t nil t)

(define-sysfun-callinfo load
  1 0 nil t (:verbose :print :if-does-not-exist :external-format))

(define-sysfun-callinfo make-array
  1 0 nil t (:element-type :initial-element :initial-contents :adjustable
             :fill-pointer :displaced-to :displaced-index-offset))

(define-sysfun-callinfo (make-list make-string) 1 0 nil t (:initial-element))

(define-sysfun-callinfo (nstring-capitalize nstring-downcase nstring-upcase
                         string-capitalize string-downcase string-upcase)
  1 0 nil t (:start :end))

(define-sysfun-callinfo open
  1 0 nil t (:direction :element-type :if-exists :if-does-not-exist
             :external-format))

(define-sysfun-callinfo parse-integer
  1 0 nil t (:start :end :radix :junk-allowed))

(define-sysfun-callinfo read-sequence 1 0 nil t (:start :end))

(define-sysfun-callinfo (write write-to-string)
  1 0 nil t (:array :base :case :circle :escape :gensym :length :level :lines
             :miser-width :pprint-dispatch :pretty :radix :readably
             :right-margin :stream))

(define-sysfun-callinfo compile-file-pathname 1 0 nil t (:output-file) t)

(define-sysfun-callinfo (delete-duplicates remove-duplicates)
  1 0 nil t (:from-end :test :test-not :start :end :key))

(define-sysfun-callinfo
  (- / apply aref array-in-bounds-p array-row-major-index concatenate
   error every invoke-restart make-condition mapc mapcan mapcar mapcon
   mapl maplist notany notevery signal some warn xlisp::%make-struct)
  1 0 t)

;;**** decode-universal-time enough-namestring ensure-directories-exist
;;**** float-sign get-dispatch-macro-character upgraded-array-element-type
;;**** upgraded-complex-part-type
(define-sysfun-callinfo
  (apropos apropos-list atan butlast ceiling compile copy-symbol
   describe digit-char digit-char-p export fceiling ffloor file-position
   find-restart find-symbol float floor fround ftruncate function-information 
   get-macro-character get-setf-expansion import intern last log logandc1
   logandc2 lognand lognor logorc1 logorc2 macro-function macroexpand
   macroexpand-1 parse-namestring pathname-device pathname-directory
   pathname-host pathname-name pathname-type pathname-version pprint
   prin1 princ print random require round shadow shadowing-import
   truncate unexport unintern unread-char unuse-package use-package
   variable-information wild-pathname-p write-char xlisp::slot-value)
  1 1)

(define-sysfun-callinfo (write-line write-string) 1 1 nil t (:start :end))

;;**** make-dispatch-macro-character read-delimited-list set-syntax-from-char
(define-sysfun-callinfo 
  (make-package make-string-input-stream merge-pathnames read-byte
   xlisp::mark-as-special)
  1 2)

(define-sysfun-callinfo parse-namestring 1 2 nil t (:start :end :junk-allowed))

(define-sysfun-callinfo read-from-string
  1 2 nil t (:start :end :preserve-whitespace))
;;**** file-string-length revappend scale-float vector-push
(define-sysfun-callinfo
  (xlisp::%set-car xlisp::%set-cdr array-dimension ash char coerce cons
   documentation elt eq eql equal equalp expt get-properties ldiff logbitp
   logtest maphash mod nreconc nth nthcdr pathname-match-p rem remhash remprop
   rename-file row-major-aref rplaca rplacd schar set string-left-trim
   string-right-trim string-trim svref tailp write-byte xlisp::%struct-ref
   xlisp::%struct-type-p xlisp::push-condition-handlers
   xlisp::apply-arg-rotate xlisp::default-structure-slot-value
   xlisp::destructure-check-arg xlisp::install-sharp-s-constructor)
  2)

(define-sysfun-callinfo (cerror map map-into) 2 0 t)

;;**** assoc-if assoc-if-not member-if-not rassoc-if rassoc-if-not
;;**** subst-if subst-if-not
(define-sysfun-callinfo (member-if sort stable-sort) 2 0 nil t (:key))

;;**** rassoc
(define-sysfun-callinfo
  (adjoin assoc intersection member nintersection nset-difference
   nset-exclusive-or nsublis nunion set-difference set-exclusive-or sublis
   subsetp union)
  2 0 nil t (:key :test :test-not))

(define-sysfun-callinfo (format) 2 0 t)

(define-sysfun-callinfo (count find position)
  2 0 nil t (:from-end :start :end :key :test :test-not))

(define-sysfun-callinfo (count-if count-if-not
                         find-if find-if-not
                         position-if position-if-not)
  2 0 nil t (:from-end :start :end :key))

(define-sysfun-callinfo (delete remove)
  2 0 nil t (:from-end :test :test-not :start :end :count :key))

(define-sysfun-callinfo (delete-if delete-if-not remove-if remove-if-not)
  2 0 nil t (:from-end :start :end :count :key))

(define-sysfun-callinfo fill 2 0 nil t (:start :end))

;;**** (define-sysfun-callinfo make-sequence 2 0 nil t (:initial-element))

(define-sysfun-callinfo xlisp::install-function 2 2)

;;**** (define-sysfun-callinfo merge 4 0 nil t (:key))

;;**** (define-sysfun-callinfo (mismatch search)
;;****  2 0 nil t (:from-end :test :test-not :key :start1 :start2 :end1 :end2))

(define-sysfun-callinfo reduce
  2 0 nil t (:key :from-end :start :end :initial-value)) 

(define-sysfun-callinfo
  (replace string-equal string-greaterp string-lessp string-not-equal
   string-not-greaterp string-not-lessp string/= string< string<= string=
   string> string>= )
  2 0 nil t (:start1 :end1 :start2 :end2))

;;**** (define-sysfun-callinfo tree-equal 2 0 nil t (:test :test-not))

(define-sysfun-callinfo write-sequence 2 0 nil t (:start :end))

;;**** vector-push-extend
(define-sysfun-callinfo (get getf gethash pairlis rename-package subseq
                         subtypep typep)
  2 1)

(define-sysfun-callinfo set-macro-character 2 2)
(define-sysfun-callinfo
  (xlisp::%set-elt xlisp::%set-get xlisp::%set-gethash xlisp::%set-nth
   xlisp::%set-svref acons xlisp::%set-rm-aref xlisp::%struct-set
   xlisp::make-macro-lambda-expression xlisp::install-structure-slots)
  3)

(define-sysfun-callinfo (nsubst subst)
  3 0 nil t (:key :test :test-not))

(define-sysfun-callinfo (nsubst-if nsubst-if-not)
  3 0 nil t (:key))

(define-sysfun-callinfo (nsubstitute substitute)
  3 0 nil t (:from-end :test :test-not :start :end :count :key))

(define-sysfun-callinfo (nsubstitute-if nsubstitute-if-not
                         substitute-if substitute-if-not)
  3 0 nil t (:from-end :start :end :count :key))

;;**** (define-sysfun-callinfo translate-pathname 3 0 nil t nil t)

(define-sysfun-callinfo xlisp::%set-aref 3 0 t)

;;**** set-dispatch-macro-character
(define-sysfun-callinfo (parse-macro) 3 1)

#+xlisp-stat
(define-sysfun-callinfo xlisp::add-method 4)

;;**** (define-sysfun-callinfo encode-universal-time 6)
