(in-package "XLISP")

;;;
;;; Global Variables
;;;

(defvar *inline-table* (make-hash-table))
(defvar *global-environment* nil)


;;;
;;; PROCLAIM, and DECLAIM
;;;
(export '(proclaim declaim))

(defun proclaim (decl)
  (case (first decl)
        (special (dolist (s (rest decl)) (mark-as-special s)))
        (inline (dolist (s (rest decl)) (proclaim-inline s t)))
        (notinline (dolist (s (rest decl)) (proclaim-inline s nil)))))

(defmacro declaim (&rest args)
  (when args
        `(eval-when (:compile-toplevel :load-toplevel :execute)
                    ,@(mapcar #'(lambda (x) `(proclaim ',x)) args))))

(defun proclaim-inline (s inline)
  (let ((val (if inline 'inline 'notinline))
        (entry (gethash s *inline-table*)))
    (if entry
        (setf (first entry) val)
      (setf (gethash s *inline-table*) (list val nil)))))



;;;
;;; AUGMENT-ENVIRONMENT, VARIABLE-INFORMATION, and FUNCTION-INFORMATION
;;;

(export '(augment-environment variable-information function-information))

(defun augment-environment (env &key
                                variable symbol-macro function macro declare)
  (when (intersection variable (mapcar #'first symbol-macro))
        (error "adding some symbols as variables and as symbol macros"))
  (when (intersection function (mapcar #'first macro))
        (error "adding some symbols as functions and macros"))
  (setf env (augment-environment-variable variable env))
  (setf env (augment-environment-symbol-macro symbol-macro env))
  (setf env (augment-environment-function function env))
  (setf env (augment-environment-macro macro env))
  (setf env (augment-environment-declare declare env))
  env)

(defun augment-environment-variable (vars env)
  (dolist (v vars)
    (when (special-variable-p v)
          (error "can't make lexical binding for the special variable ~s" v)))
  (if vars
      (let ((frame (mapcar #'list vars)))
        (cons (cons frame (car env)) (cdr env)))
    env))

(defun special-variable-p (sym)
  (or (specialp sym)
      (first (get-vdecls sym *global-environment*))
      (and (boundp '*cmp-specials*) (member sym *cmp-specials*))))

(defun augment-environment-symbol-macro (smacs env)
  (dolist (sm smacs)
    (let ((v (first sm)))
      (when (special-variable-p v)
            (error "can't make lexical binding for the special variable ~s"
                   v))))
  (if smacs
      (let ((frame nil))
        (dolist (s smacs)
          (push (cons :symbol-macro (second s)) frame)
          (push (cons (first s) :symbol-macro) frame))
        (cons (cons frame (car env)) (cdr env)))
    env))

(defun augment-environment-function (funs env)
  (if funs
      (cons (car env) (cons (mapcar #'list funs) (cdr env)))
    env))

(defun augment-environment-macro (macs env)
  (flet ((mkmac (x) (cons (first x) (coerce-to-macro (second x)))))
    (if macs
        (cons (car env) (cons (mapcar #'mkmac macs) (cdr env)))
      env)))

(defun augment-environment-declare (decls env)
  (setf env (cons (cons nil (car env)) (cons nil (cdr env))))
  (dolist (d decls)
    (case (first d)
          (special (add-special-decls (rest d) env))
          ((ignore ignorable) (add-ignore-decls (first d) (rest d) env))
          (type (add-type-decls (second d) (rest (rest d)) env))
          ((array atom bignum bit bit-vector character compiled-function
            complex cons double-float fixnum float function hash-table integer
            keyword list long-float nil null number package pathname
            random-state ratio rational readtable sequence short-float
            signed-byte simple-array simple-bit-vector simple-string
            simple-vector single-float standard-char stream string symbol t
            unsigned-byte vector extended-character real)
           (add-type-decls (first d) (rest d) env))
          ((inline notinline) (add-inline-decls (first d) (rest d) env))
          (ftype (add-ftype-decls (second d) (rest (rest d)) env))))
  (when (null (car (car env))) (pop (car env)))
  (when (null (car (cdr env))) (pop (cdr env)))
  (if (or (car env) (cdr env)) env nil))

(defun get-decls-frame (v frame dflt)
  (dolist (entry frame)
    (cond
     ((and (eq (first entry) :declare) (eq (second entry) v))
      (return (rest (rest entry))))
     ((eq (first entry) v) (return dflt)))))

(defun get-vdecls (v env)
  (dolist (frame (car env) '(nil nil t))
    (let ((val (get-decls-frame v frame '(nil nil t))))
      (when val (return val)))))

(defun get-fdecls (v env)
  (dolist (frame (cdr env) (list (first (gethash v *inline-table*)) t))
    (let ((val (get-decls-frame v frame '(nil t))))
      (when val (return val)))))

(defun add-special-decls (vars env)
  (dolist (v vars)
    (let ((decls (get-vdecls v env)))
      (push `(:declare ,v t ,@(rest decls)) (car (car env))))))

(defun add-type-decls (type vars env)
  (dolist (v vars)
    (let* ((decls (get-vdecls v env))
           (ntype (combine-types (third decls) type)))
      (push `(:declare ,v ,(first decls) ,(second decls) ,ntype)
            (car (car env))))))

(defun combine-types (a b)
  (flet ((wrap (x) (if (and (consp x) (eq (first x) 'and)) (rest x) (list x))))
    (cond
     ((eq a t) b)
     ((eq b t) a)
     (t (cons 'and (append (wrap a) (wrap b)))))))

(defun add-ignore-decls (which vars env)
  (dolist (v vars)
    (let* ((decls (get-vdecls v env)))
      (push `(:declare ,v ,(first decls) ,which ,(third decls))
            (car (car env))))))

(defun add-ftype-decls (type vars env)
  (dolist (v vars)
    (let* ((decls (get-fdecls v env))
           (ntype (combine-types (second decls) type)))
      (push `(:declare ,v ,(first decls) ,ntype) (car (cdr env))))))

(defun add-inline-decls (which vars env)
  (dolist (v vars)
    (let ((decls (get-fdecls v env)))
      (push `(:declare ,v ,which ,(second decls)) (car (cdr env))))))

(defun variable-information (sym &optional (env *global-environment*))
  (multiple-value-bind (found value extra)
                       (get-environment-entry sym (first env))
   (let* ((dinfo (get-vdecls sym env))
          (special (or (first dinfo) (special-variable-p sym)))
          (decls (let ((decls nil))
                   (unless (eq (third dinfo) t)
                           (push `(type ,(third dinfo)) decls))
                   (when (second dinfo) (push `(ignore ,(second dinfo)) decls))
                   decls)))
     (cond
      (found
       (if (and (eq value :symbol-macro) (eq (car extra) :symbol-macro))
           (values :symbol-macro t decls)
         (values (if special :special :lexical) t decls)))
      ((constantp sym) (values :constant nil decls))
      (special (values :special nil decls))
      ((nth-value 1 (global-symbol-macro sym))
       (values :symbol-macro nil decls))
      (t (values nil nil decls))))))

(defun function-information (sym &optional (env *global-environment*))
  (flet ((macrop (f) (and (or (typep f 'byte-code-closure) (typep f 'closure))
                          (not (functionp f))))
         (global-frame-p (f) (or (member f (cdr *global-environment*))
                                 (eq f *cmp-global-macros*)
                                 (and (boundp '*cmp-macros*)
                                      (eq f *cmp-macros*)))))
    (multiple-value-bind (found value extra frame)
                         (get-environment-entry sym (rest env))
      (declare (ignore extra))
      (let* ((dinfo (get-fdecls sym env))
             (decls (let ((decls nil)
                          (inline (first dinfo))
                          (type (second dinfo)))
                      (unless (member type '(function t))
                              (push `(ftype ,type) decls))
                      (when inline (push `(inline ,inline) decls)))))
        (cond
         (found
          (if (macrop value)
              (values :macro (not (global-frame-p frame)) nil)
            (values :function t decls)))
         ((fboundp sym)
          (let ((fun (symbol-function sym)))
            (cond
             ((macrop fun) (values :macro nil nil))
             ((typep fun 'fsubr) (values :special-operator nil decls))
             (t (values (if (functionp fun) :function nil) nil decls)))))
         (t (values nil nil decls)))))))


;;;
;;; Modified Functions and Macros
;;;

(defun split-declarations (x)
  (flet ((head-is-declaration (x)
           (and (consp (first x)) (eq (first (first x)) 'declare)))
         (head-is-docstring (x) (and (stringp (first x)) (consp (rest x)))))
    (do ((decls nil)
         (body x (rest body))
         (doc nil))
        (nil)
        (cond
         ((head-is-declaration body) (push (first body) decls))
         ((head-is-docstring body) (setf doc (first body)))
         (t (return (list (nreverse decls) body doc)))))))

(define-special-form-macro defvar (var &optional (value nil have-val) doc)
  `(progn 
     (eval-when (:compile-toplevel)
                (when (boundp '*cmp-specials*)
                      (pushnew ',var *cmp-specials*))
                (setf *global-environment*
                      (augment-environment *global-environment*
                                           :declare '((special ,var)))))
     (mark-as-special ',var)
     ,@(if have-val `((unless (boundp ',var) (set ',var ,value))))
     ,@(if doc `((%set-get ',var 'variable-documentation ,doc)))
     ',var))

(define-special-form-macro defparameter (var value &optional doc)
  `(progn
     (eval-when (:compile-toplevel)
                (when (boundp '*cmp-specials*)
                      (pushnew ',var *cmp-specials*))
                (setf *global-environment*
                      (augment-environment *global-environment*
                                           :declare '((special ,var)))))
     (mark-as-special ',var)
     (set ',var ,value)
     ,@(if doc `((%set-get ',var 'variable-documentation ,doc)))
     ',var))
(define-special-form-macro defmacro (f args &rest body &environment env)
  (multiple-value-bind (fexpr doc)
                       (make-macro-lambda-expression f args body)
    `(progn
       (eval-when (:compile-toplevel)
                  (push (cons ',f (coerce-to-macro ,fexpr)) *cmp-macros*)
                  ,(let ((macro (parse-macro f args body env)))
                     `(setf *global-environment*
                            (augment-environment *global-environment*
                                                 :macro '((,f ,macro)))))
)
       ,@(if doc `((%set-get ',f 'function-documentation ,doc)))
       (install-function ',f (coerce-to-macro ,fexpr)))))
#+xlisp-stat
(define-special-form-macro defmeth (ob sym args &rest body &environment env)
  (let* ((db (split-declarations body))
         (decls (cons '(declare (ignorable self)) (first db)))
         (b (second db))
         (doc (third db))
         (fexpr `#'(lambda ,(cons 'self args) ,@decls (block ,sym ,@b)))
         (obform (if (and (symbolp ob)
                          (not (eq (variable-information ob env) :lexical)))
                     `(symbol-value ',ob)
                   ob)))
    `(progn
       (xlisp::add-method ,obform ',sym ,fexpr ,doc)
       ',sym)))
#+xlisp-stat
(defmacro def (symbol value &environment env)
"Syntax: (def var form)
VAR is not evaluated and must be a symbol.  Assigns the value of FORM to
VAR and adds VAR to the list *VARIABLES* of def'ed variables. Returns VAR.
If VAR is already bound and the global variable *ASK-ON-REDEFINE*
is not nil then you are asked if you want to redefine the variable."
  (cond
   ((eq (variable-information symbol env) :lexical)
    (warn "DEF should not be used for the local variable ~s" symbol)
    `(progn (setf ,symbol ,value) ',symbol))
   (t `(unless (and *ask-on-redefine*
                    (boundp ',symbol)
                    (not (y-or-n-p "Variable has a value. Redefine?")))
               (pushnew ',symbol *variables*)
               (setf (symbol-value ',symbol) ,value)
               ',symbol))))
