(in-package "NEWCMP")

;;;;
;;;; Splitting Functions
;;;;

(defstruct (spfentry (:type list) (:constructor make-spfentry (name def)))
  name def closure local tail)

(defun lookup-split-functions-def (f env)
  (let ((entry (assoc f env)))
    (if entry (spfentry-def entry) nil)))

(macrolet ((define-lookup (name which)
             `(defun ,name (f env)
                (let ((entry (assoc f env)))
                  (if entry
                      (let ((val (,which entry)))
                        (if val
                            val
                          (let ((nval (spfentry-var entry)))
                            (setf (,which entry) nval)
                            nval)))
                    f)))))
  (define-lookup lookup-split-functions-closure spfentry-closure)
  (define-lookup lookup-split-functions-local spfentry-local)
  (define-lookup lookup-split-functions-tail spfentry-tail))

(defun spfentry-var (entry)
  (let ((name (spfentry-name entry)))
    (if (or (spfentry-closure entry)
            (spfentry-local entry)
            (spfentry-tail entry))
        (copy-var name)
      name)))

(defun split-all-functions (e) (split-functions e nil t))

(defun split-local-functions (e) (split-functions e nil nil))

(defun split-functions (e env full)
  (cond
   ((letrec-p e) (split-functions-letrec e env full))
   ((call-p e) (split-functions-call e env full))
   (t (split-functions-cexpr e env full))))

(defun split-functions-letrec (e env full)
  (let* ((vars (letrec-vars e))
         (args (letrec-args e))
         (frame (mapcar #'make-spfentry vars args))
         (nenv (append frame env))
         (sargs (mapcar #'(lambda (x) (split-functions x nenv full)) args))
         (nbody (split-functions (letrec-body e) nenv full)))
    (do ((frame frame (rest frame))
         (sargs sargs (rest sargs)))
        ((or (null frame) (null sargs)))
        (setf (spfentry-def (first frame)) (first sargs)))
    (let ((nargs nil)
          (nvars nil))
      (dolist (entry frame)
        (let ((def (spfentry-def entry))
              (closure (spfentry-closure entry))
              (local (spfentry-local entry))
              (tail (spfentry-tail entry)))
          (when (and full (null tail) closure local)
                (let ((var (copy-var (spfentry-name entry))))
                  (setf (spfentry-tail entry) var) ;;**** is this needed?
                  (setf tail var)))
          (when tail
                (push tail nvars)
                (push (fixup-split-function def 'tail-lambda) nargs))
          (when closure
                (push closure nvars)
                (let ((other (if full tail local)))
                  (if other
                      (push (wrap-split-function-tail def other 'lambda) nargs)
                    (let ((ndef (copy-func def)))
                      (setf (func-type ndef) 'lambda)
                      (push ndef nargs)))))
          (when local
                (push local nvars)
                (let ((ndef (fixup-split-function def 'local-lambda)))
                  (push (if tail
                            (wrap-split-function-tail ndef tail 'local-lambda)
                          ndef)
                        nargs)))))
      (remake-cexpr e nil (nreverse nargs) (nreverse nvars) (list nbody)))))

(defun fixup-split-function (def newtype)
  (let ((nopt (func-nopt def))
        (has-rest (func-has-rest def))
        (type (func-type def)))
    (if (and (= nopt 0) (not has-rest) (eq type newtype))
        def
      (let ((ndef (copy-func def)))
        (incf (func-nreq ndef) (if has-rest (+ nopt 1) nopt))
        (setf (func-nopt ndef) 0)
        (setf (func-has-rest ndef) nil)
        (setf (func-type ndef) newtype)
        ndef))))

(defun wrap-split-function-tail (def fun type)
  (let ((ndef (copy-func def))
        (ctx (func-context def))
        (nvars (mapcar #'copy-var (func-vars def))))
    (setf (func-type ndef) type)
    (setf (func-vars ndef) nvars)
    (setf (func-conts ndef) (list (make-call fun nvars nil nil nil nil ctx)))
    ndef))

(defun split-functions-call (e env full)
  (let* ((nargs (mapcar #'(lambda (f) (lookup-split-functions-closure f env))
                        (call-args e)))
         (form (call-form e))
         (nconts (if form (list (split-functions form env full)) nil))
         (fun (if (and full (null nconts))
                  (lookup-split-functions-tail (call-fun e) env)
                (lookup-split-functions-local (call-fun e) env)))
         (def (lookup-split-functions-def (call-fun e) env))
         (ne (remake-cexpr e fun nargs (cexpr-vars e) nconts)))
    (if def
        (if (call-is-mvcall e)
            (rewrite-split-functions-mvcall ne def)
          (rewrite-split-functions-call ne def))
      ne)))

(defun rewrite-split-functions-mvcall (e def)
  (let* ((nreq (func-nreq def))
         (nopt (func-nopt def))
         (has-rest (func-has-rest def))
         (name (func-name def))
         (nvars (mapcar #'copy-var (func-vars def)))
         (ne (copy-call e))
         (data (list nreq nopt has-rest name)))
    (setf (call-is-mvcall ne) nil)
    (setf (call-args ne) nvars)
    (make-primop 'load-mvargs nil nvars ne data (call-context ne))))

(defun rewrite-split-functions-call (e def)
  (let* ((nreq (func-nreq def))
         (nopt (func-nopt def))
         (has-rest (func-has-rest def))
         (nro (+ nreq nopt))
         (args (call-args e))
         (n (length args)))
    (unless (<= nreq n) (too-few-arguments e nreq))
    (unless (or (<= n nro) has-rest) (too-many-arguments e nro))
    (if (or (< n nro) has-rest)
        (let* ((ov (if (< n nro) (make-tmp-var) nil))
               (rv (if has-rest
                       (if (< nro n) (make-tmp-var) (make-const nil))
                     nil))
               (oargs (if ov (make-list (- nro n) :initial-element ov) nil))
               (rargs (if has-rest (list rv) nil))
               (tail (nthcdr nro args))
               (head (ldiff args tail))
               (nargs (append head oargs rargs))
               (ne (copy-call e))
               (ctx (primop-context ne)))
          (setf (call-args ne) nargs)
          (when rv (setf ne (wrap-split-functions-rest ne rv tail ctx)))
          (when ov
                (setf ne (make-primop 'load-missing nil (list ov) ne nil ctx)))
          ne)
      e)))

(defun wrap-split-functions-rest (e v args ctx)
  (if args
      (do ((args args (rest args)))
          ((null (rest args))
           (make-primop 'cons (list (first args) (make-const nil)) (list v) e
                        nil ctx))
          (let ((av (make-tmp-var)))
            (setf e (make-primop 'cons (list (first args) av) (list v) e
                                 nil ctx))
            (setf v av)))
    e))

(defun split-functions-cexpr (e env full)
  (let ((nargs (mapcar #'(lambda (f) (lookup-split-functions-closure f env))
                       (cexpr-args e)))
        (nconts (mapcar #'(lambda (x) (split-functions x env full))
                        (cexpr-conts e))))
    (remake-cexpr e (cexpr-fun e) nargs (cexpr-vars e) nconts)))

;;;;
;;;; Finding Free Variables
;;;;

(defstruct (freemap-entry (:type list)
                          (:constructor
                           (make-freemap-entry (def &optional free))))
  def free bound)

(defun make-freemap () (make-hash-table))

(defun add-freemap-entry (entry freemap)
  (setf (gethash (freemap-entry-def entry) freemap) entry))

(defun lookup-freemap-entry (e freemap)
  (gethash e freemap))

(defun init-freemap (e map entry lfuns)
  (if (letrec-p e)
      (init-freemap-letrec e map entry lfuns)
    (init-freemap-cexpr e map entry lfuns)))

(defun init-freemap-letrec (e map entry lfuns)
  (let* ((vars (cexpr-vars e))
         (nlfuns (append vars lfuns))
         (afree nil))
    (dolist (a (letrec-args e))
      (let ((ae (make-freemap-entry a)))
        (add-freemap-entry ae map)
        (init-freemap a map ae nlfuns)
        (setf afree (union (freemap-entry-free ae) afree))))
    (init-freemap (letrec-body e) map entry nlfuns)
    (let ((free (union afree (freemap-entry-free entry)))
          (bound (freemap-entry-bound entry)))
      (setf (freemap-entry-free entry) (set-difference free vars))
      (setf (freemap-entry-bound entry) (union vars bound)))))

(defun init-freemap-cexpr (e map entry lfuns)
  (let ((vars (cexpr-vars e))
        (args (cexpr-args e))
        (conts (cexpr-conts e))
        (fun (cexpr-fun e)))
    (dolist (c conts)
      (init-freemap c map entry lfuns))
    (let ((free (freemap-entry-free entry))
          (bound (freemap-entry-bound entry)))
      (when (and (var-p fun) (not (member fun lfuns)))
            (pushnew fun free))
      (dolist (a args)
        (when (var-p a) (pushnew a free)))
      (setf (freemap-entry-free entry) (set-difference free vars))
      (setf (freemap-entry-bound entry) (union vars bound)))))

(defun freevars (e)
  (let* ((gr (compute-callgr e))
         (funs (callgr-functions gr))
         (freemap (make-freemap))
         (entry (make-freemap-entry e)))
    (add-freemap-entry entry freemap)
    (init-freemap e freemap entry nil)
    (loop
     (let ((changed nil))
       (dolist (fun funs)
         (let ((entry (lookup-freemap-entry fun freemap)))
           (let* ((free (freemap-entry-free entry))
                  (bound (freemap-entry-bound entry)))
             (dolist (c (callgr-callees (first entry) gr))
               (dolist (v (freemap-entry-free (lookup-freemap-entry c freemap)))
                 (unless (or (member v free) (member v bound))
                         (push v free)
                         (setf changed t))))
             (setf (freemap-entry-free entry) free))))
       (unless changed (return))))
    freemap))


;;;;
;;;; Lambda Lifting
;;;;

(defun lambda-lift (e) (lambda-lift-cexpr e nil (freevars e)))

(defun lambda-lift-cexpr (e env map)
  (cond
   ((func-p e) (lambda-lift-func e env map))
   ((letrec-p e) (lambda-lift-letrec e env map))
   ((call-p e) (lambda-lift-call e env map))
   (t (let ((nargs (sublis env (cexpr-args e)))
            (nconts (mapcar #'(lambda (x) (lambda-lift-cexpr x env map))
                            (cexpr-conts e))))
        (remake-cexpr e (cexpr-fun e) nargs (cexpr-vars e) nconts)))))

(defun lambda-lift-func (e env map)
  (let* ((entry (lookup-freemap-entry e map))
         (free (if entry (freemap-entry-free entry) nil))
         (fvars (mapcar #'copy-var free))
         (nenv (append (mapcar #'cons free fvars) env))
         (nform (lambda-lift-cexpr (func-body e) nenv map))
         (nvars (append fvars (func-vars e)))
         (ne (remake-cexpr e nil nil nvars (list nform)))
         (nfree (length fvars)))
    (when fvars
          (incf (func-nreq ne) nfree)
          (setf (func-nfree ne) nfree))
    ne))

(defun lambda-lift-letrec (e env map)
  (let ((args (letrec-args e))
        (vars (letrec-vars e)))
    (multiple-value-bind (nenv cvars clabels cfuns cfree)
                         (lambda-lift-letrec-info args vars env map)
      (let* ((nbody (lambda-lift-cexpr (letrec-body e) nenv map))
             (nargs (mapcar #'(lambda (a) (lambda-lift-func a nenv map)) args))
             (ctx (primop-context e))
             (nnbody (wrap-set-closure-code cvars clabels cfuns nbody ctx))
             (ne (make-letrec vars nargs nnbody ctx)))
        (if cvars
            (if cfree
                (let* ((scargs (cons (first cvars) cfree))
                       (nne (make-primop 'set-closure-data scargs nil ne
                                         nil (primop-context ne))))
                  (make-primop 'make-closures nil cvars nne (length cfree)
                               ctx))
              (make-primop 'make-closures nil cvars ne 0 ctx))
          ne)))))

(defun lambda-lift-letrec-info (args vars env map)
  (let ((nenv env)
        (cvars nil)
        (clabels nil)
        (cfuns nil)
        (cfree nil))
    (do ((vars vars (rest vars))
         (args args (rest args)))
        ((or (null vars) (null args)))
        (let* ((v (first vars))
               (a (first args))
               (free (freemap-entry-free (lookup-freemap-entry a map))))
          (if (eq (func-type a) 'lambda)
              (let ((cv (copy-var v)))
                (push cv cvars)
                (push v clabels)
                (push a cfuns)
                (push (cons v cv) nenv)
                (setf cfree (union free cfree)))
            (add-freemap-entry (make-freemap-entry v free) map))))
    (dolist (c cfuns) (add-freemap-entry (make-freemap-entry c cfree) map))
    (let ((ncfree (sublis nenv cfree))) ;; in case closures are free
      (values nenv cvars clabels cfuns ncfree))))

(defun wrap-set-closure-code (cvars clabels cfuns body ctx)
  (do ((cvars cvars (rest cvars))
       (clabels clabels (rest clabels))
       (cfuns cfuns (rest cfuns)))
      ((or (null cvars) (null clabels) (null cfuns)))
      (let* ((cv (first cvars))
             (lbl (first clabels))
             (name (func-name (first cfuns)))
             (iname (if (symbol-package name) name nil)))
        (setf body
              (make-primop 'set-closure-code (list cv lbl) nil body iname
                           ctx))))
  body)

(defun lambda-lift-call (e env map)
  (let* ((fun (call-fun e))
         (nfun (sublis env fun))
         (args (call-args e))
         (vars (call-vars e))
         (form (call-form e))
         (nconts (if form (list (lambda-lift-cexpr form env map)) nil))
         (entry (lookup-freemap-entry fun map))
         (free (if entry (freemap-entry-free entry) nil))
         (nargs (sublis env (append free args))))
    (remake-cexpr e nfun nargs vars nconts)))


;;;;
;;;; Constant Lifting
;;;;

(defun lift-constants (e) (lift-constants-1 e nil))

(defun lift-constants-1 (e env)
  (flet ((lift-env (x) (lift-constants-1 x env)))
    (let ((fun (let ((fun (cexpr-fun e)))
                 (if (and (const-p fun) (symbolp (const-value fun)))
                     (const-value fun)
                   fun)))
          (args (cexpr-args e))
          (vars (cexpr-vars e)))
      (if (some #'const-p args)
          (let ((rargs nil)
                (cvars nil)
                (consts nil))
            (dolist (a args)
              (if (const-p a)
                  (let ((v (cdr (flet ((test (x y) (eql (const-value x) (const-value y))))
                                  (assoc a env :test #'test)))))
                    (if v
                        (push v rargs)
                      (let* ((sym (gensym "CONST"))
                             (v (make-var sym `(eql ,(const-value a)) nil)))
                        (push v cvars)
                        (push v rargs)
                        (push a consts)
                        (push (cons a v) env))))
                (push a rargs)))
            (let* ((nconts (mapcar #'lift-env (cexpr-conts e)))
                   (ne (remake-cexpr e fun (nreverse rargs) vars nconts))
                   (ctx (cexpr-context ne)))
              (do ((cvars cvars (rest cvars))
                   (consts consts (rest consts)))
                  ((or (null cvars) (null consts)))
                  (let ((vl (list (first cvars)))
                        (cv (const-value (first consts))))
                    (setf ne (make-primop 'load-constant nil vl ne cv ctx))))
              ne))
        (let ((nconts (mapcar #'lift-env (cexpr-conts e)))
              (nargs (if (letrec-p e) (mapcar #'lift-constants args) args)))
          (remake-cexpr e fun nargs vars nconts))))))
