(in-package "NEWCMP")


;;;;
;;;; Drop Unused Functions
;;;;

(defun find-dead-functions (e)
  (let* ((gr (compute-callgr e t))
         (funs (callgr-functions gr))
         (live (list e))
         (follow live))
    (loop
     (let ((f (pop follow)))
       (dolist (c (callgr-callees f gr))
         (unless (member c live)
                 (push c live)
                 (pushnew c follow)))
       (unless follow (return (set-difference funs live)))))))

(defun drop-dead-functions (e)
  (let ((funs (find-dead-functions e)))
    (labels ((drop-funs (e)
               (let* ((args (cexpr-args e))
                      (vars (cexpr-vars e))
                      (conts (cexpr-conts e))
                      (nconts (mapcar #'drop-funs conts)))
                 (if (letrec-p e)
                     (let ((nargs nil)
                           (nvars nil))
                       (mapc #'(lambda (a v)
                                 (unless (member a funs)
                                         (push (drop-funs a) nargs)
                                         (push v nvars)))
                             args vars)
                       (setf nargs (nreverse nargs))
                       (setf nvars (nreverse nvars))
                       (if (null nargs)
                           (first nconts)
                         (remake-cexpr e nil nargs nvars nconts)))
                   (remake-cexpr e (cexpr-fun e) args vars nconts)))))
      (if funs (drop-funs e) e))))



;;;;
;;;; Substituting Functions Called Only Once
;;;;


;;;
;;; Counting Function Uses
;;;

(defstruct (usecounts-entry (:type list)
                            (:constructor make-usecounts-entry (def)))
  def (calls 0) (args 0))

(defun add-usecounts-entry (name def table)
  (setf (gethash name table) (make-usecounts-entry def)))

(defun function-counts (e)
  (let ((table (make-hash-table)))
    (fill-usecounts-table e table)
    table))

(defun fill-usecounts-table (e table)
  (cond
   ((letrec-p e) (fill-usecounts-table-letrec e table))
   ((call-p e) (fill-usecounts-table-call e table))
   (t (fill-usecounts-table-cexpr e table))))

(defun fill-usecounts-table-letrec (e table)
  (let ((args (letrec-args e))
        (vars (letrec-vars e)))
    (do ((args args (rest args))
         (vars vars (rest vars)))
        ((or (null args) (null vars)))
        (add-usecounts-entry (first vars) (first args) table))
    (dolist (a args) (fill-usecounts-table a table))
    (fill-usecounts-table (letrec-body e) table)))

(defun fill-usecounts-table-call (e table)
  (let ((entry (gethash (call-fun e) table)))
    (when entry (incf (usecounts-entry-calls entry))))
  (fill-usecounts-table-cexpr e table))

(defun fill-usecounts-table-cexpr (e table)
  (dolist (v (cexpr-args e))
    (let ((entry (gethash v table)))
      (when entry (incf (usecounts-entry-args entry)))))
  (dolist (x (cexpr-conts e)) (fill-usecounts-table x table)))

(defun find-once-called-functions (e)
  (let ((table (function-counts e))
        (env nil))
    (maphash #'(lambda (name entry)
                 (when (and (= (usecounts-entry-calls entry) 1)
                            (= (usecounts-entry-args entry) 0))
                       (push (cons name (usecounts-entry-def entry)) env)))
             table)
    env))


;;;
;;; Substituting the Functions
;;;

(defun substitute-once-called-functions (e)
  (let ((e (split-local-functions e)))
    (loop
     (let* ((ne (drop-dead-functions e))
            (env (find-once-called-functions ne)))
       (when env (setf ne (substitute-values (substitute-once-called ne env))))
       (if (eq e ne)
           (return e)
         (setf e ne))))))

(defun substitute-once-called (e env)
  (cond
   ((call-p e) (substitute-once-called-call e env))
   ((letrec-p e) (substitute-once-called-letrec e env))
   (t (substitute-once-called-cexpr e env))))

(defun substitute-once-called-call (e env)
  (let ((def (cdr (assoc (call-fun e) env))))
    (if def
        (let* ((ndef (substitute-once-called def env))
               (vars (call-vars e))
               (vi (call-val-ignored e))
               (ctx (call-context e))
               (form (call-form e))
               (kv (if form (if vars (first vars) (if vi nil t)) nil))
               (kf (if form (substitute-once-called form env) nil)))
          (substitute-call ndef (call-args e) kv kf ctx))
      (substitute-once-called-cexpr e env))))

(defun substitute-once-called-letrec (e env)
  (let ((nargs nil)
        (nvars nil))
    (mapc #'(lambda (a v)
              (unless (assoc v env)
                      (push (substitute-once-called a env) nargs)
                      (push v nvars)))
          (letrec-args e) (letrec-vars e))
    (let ((nbody (substitute-once-called (letrec-body e) env)))
      (if nvars
          (remake-cexpr e nil (nreverse nargs) (nreverse nvars) (list nbody))
        nbody))))

(defun substitute-once-called-cexpr (e env)
  (let ((nconts (mapcar #'(lambda (x) (substitute-once-called x env))
                        (cexpr-conts e))))
    (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts)))


;;;
;;; Substituting a Call
;;;

(defun substitute-call (def args kv kf ctx)
  (let* ((fbody (func-body def))
         (bbody (if kf
                    (let* ((fvars (if (var-p kv) (list kv) nil))
                           (kfun (make-func nil fvars kf 'lambda ctx))
                           (k (make-tmp-var))
                           (body (substitute-continuation fbody kv k nil)))
                      (make-letrec (list k) (list kfun) body ctx))
                  fbody)))
    (make-bind (func-vars def) args bbody ctx)))

(defun substitute-continuation (e kv k funs)
  (let ((conts (cexpr-conts e)))
    (if conts
        (let ((nconts (mapcar #'(lambda (x)
                                  (substitute-continuation x kv k funs))
                              conts)))
          (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts))
      (cond
       ((primop-p e) (substitute-continuation-primop e kv k))
       ((call-p e) (if (member (call-fun e) funs)
                       e
                     (substitute-continuation-call e kv k)))
       (t (error "expression of type ~s should not be in tail position"
                 (type-of e)))))))

(defun substitute-continuation-primop (e kv k)
  (let ((ctx (primop-context e))
        (fun (primop-fun e))
        (args (primop-args e)))
    (case fun
          ((go return-from throw) e)
          (values
           (cond
            ((var-p kv)
             (let* ((args (primop-args e))
                    (nargs (list (if args (first args) (make-const nil)))))
               (make-call k nargs nil nil nil nil ctx)))
            ((null kv) (make-call k nil nil nil nil nil ctx))
            (t (remake-cexpr e fun args nil
                             (list (make-call k nil nil nil nil nil ctx))))))
          (values-list
           (cond
            ((var-p kv)
             (let ((args (primop-args e)))
               (if (null (rest args))
                   (let* ((nvars (list (copy-var kv)))
                          (nform (make-call k nvars nil nil nil nil ctx)))
                     (make-primop 'car args nvars nform nil ctx))
                 (let ((narg (if args (first args) (make-const nil))))
                   (make-call k (list narg) nil nil nil nil ctx)))))
            ((null kv) (make-call k nil nil nil nil nil ctx))
            (t (remake-cexpr e fun args nil
                             (list (make-call k nil nil nil nil nil ctx))))))
          (t (let* ((vl (if (var-p kv) (list (copy-var kv)) nil))
                    (evl (if (eq kv t) (list (make-tmp-var)) vl))
                    (kcall (make-call k vl nil nil nil nil ctx))
                    (form (if (eq kv t)
                              (make-primop 'values evl nil kcall nil ctx)
                            kcall)))
               (remake-cexpr e fun args evl (list form)))))))

(defun substitute-continuation-call (e kv k)
  (let ((fun (call-fun e))
        (args (call-args e))
        (vl (if (var-p kv) (list (copy-var kv)) nil))
        (vi (null kv))
        (ctx (call-context e)))
    (make-call fun args vl (make-call k vl nil nil nil nil ctx) vi nil ctx)))



;;;;
;;;; Eta Reduction
;;;;

(defun eta-reduce (e)
  (drop-dead-functions (eta-subst (split-local-functions e) nil)))

(defun eta-subst (e env)
  (cond
   ((letrec-p e) (eta-subst-letrec e env))
   ((call-p e) (eta-subst-call e env))
   (t (eta-subst-cexpr e env))))

(defun eta-subst-letrec (e env)
  (let ((vars (letrec-vars e))
        (args (letrec-args e)))
    (mapc #'(lambda (v def)
              (let ((f (let ((body (func-body def)))
                         (if (and (call-p body)
                                  (null (call-form body))
                                  (not (member (call-fun body) (func-vars def))))
                             (if (call-is-mvcall body)
                                 nil
                               (let ((vars (func-vars def))
                                     (args (call-args body)))
                                 (if (and (= (length args) (length vars)) (every #'eq args vars))
                                     (call-fun body)
                                   nil)))
                           nil))))
                (when f (push (cons v f) env))))
          vars args)
    (let ((nargs (mapcar #'(lambda (x) (eta-subst x env)) args))
          (nbody (eta-subst (letrec-body e) env)))
      (remake-cexpr e nil nargs vars (list nbody)))))

(defun eta-subst-call (e env)
  (let* ((nfun (sublis env (call-fun e)))
         (form (call-form e))
         (nconts (if form (list (eta-subst form env)) nil)))
    (remake-cexpr e nfun (call-args e) (call-vars e) nconts)))

(defun eta-subst-cexpr (e env)
  (let ((nconts (mapcar #'(lambda (x) (eta-subst x env)) (cexpr-conts e))))
    (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts)))


;;;;
;;;; Copying with Apha Conversion
;;;;

(defun alpha-copy (e &optional unique-name)
  (let ((mkvar (if unique-name
                   (let ((count 0))
                     #'(lambda (v)
                         (let ((cv (copy-var v))
                               (vn (intern (format nil "V-~d" count))))
                           (setf (var-name cv) vn)
                           (incf count)
                           cv)))
                 #'copy-var)))
    (alpha-cvt e nil mkvar)))

(defun alpha-cvt (e env mkvar)
  (flet ((nvar (x) (if (var-p x) (funcall mkvar x) x)))
    (let* ((nfun (sublis env (cexpr-fun e)))
           (vars (cexpr-vars e))
           (nvars (mapcar #'nvar vars))
           (nenv (append (mapcar #'cons vars nvars) env))
           (args (cexpr-args e)))
      (flet ((alpha (x) (alpha-cvt x nenv mkvar)))
        (let ((nargs (if (letrec-p e) (mapcar #'alpha args) (sublis env args)))
              (nconts (mapcar #'alpha (cexpr-conts e))))
          (remake-cexpr e nfun nargs nvars nconts))))))


;;;;
;;;; Code Simplification
;;;;



;;;
;;; Variable Use Information
;;;

(defstruct (varinfo (:type list) (:constructor make-varinfo (def)))
  def (count 0))

(defun add-varinfo (var def table)
  (let ((entry (make-varinfo def)))
    (setf (gethash var table) entry)))

(declaim (inline lookup-varinfo))
(defun lookup-varinfo (var table) (gethash var table))

(defun show-varinfo-table (table)
  (maphash #'(lambda (x y)
               (format t "Var: ~s, count=~d~%" x (varinfo-count y)))
           table))

(defun compute-varinfo-table (e)
  (let ((table (make-hash-table)))
    (fill-varinfo-table e table)
    table))

(defun fill-varinfo-table (e table)
  (flet ((incvc (x)
           (when (var-p x)
                 (incf (varinfo-count (lookup-varinfo x table))))))
    (when (call-p e) (incvc (call-fun e)))
    (dolist (v (cexpr-vars e)) (add-varinfo v e table))
    (if (letrec-p e)
        (dolist (a (cexpr-args e)) (fill-varinfo-table a table))
      (dolist (a (cexpr-args e)) (incvc a)))
    (dolist (c (cexpr-conts e)) (fill-varinfo-table c table))))


;;;
;;; Simplification Driver
;;;

(defun simplify (e)
  (substitute-values (simplify-top-down e nil (compute-varinfo-table e))))

(defun simplify-top-down (e env table)
  (multiple-value-bind (ne nenv)
                       (simplify-one-down e env table)
    (flet ((simp (e) (simplify-top-down e nenv table)))
      (let* ((fun (cexpr-fun ne))
             (nargs (let ((args (cexpr-args ne)))
                      (if (letrec-p ne) (mapcar #'simp args) args)))
             (vars (cexpr-vars ne))
             (nconts (mapcar #'simp (cexpr-conts ne)))
             (nne (remake-cexpr ne fun nargs vars nconts)))
        (dolist (v vars) (setf (varinfo-def (lookup-varinfo v table)) nne))
        (simplify-one-up nne table)))))

(defun simplify-one-down (e env table)
  (loop
   (let* ((args (cexpr-args e))
          (nargs (subvars args env table))
          (fun (cexpr-fun e))
          (nfun (subvars fun env table)))
     (unless (and (eq fun nfun)
                  (or (eq args nargs) (every #'eq args nargs)))
             (let* ((vars (cexpr-vars e))
                    (conts (cexpr-conts e))
                    (ne (remake-cexpr e nfun nargs vars conts)))
               (dolist (v vars)
                       (setf (varinfo-def (lookup-varinfo v table)) ne))
               (setf e ne))))
   (let ((ne (cond
              ((call-p e) (simplify-call e table))
              ((primop-p e) (simplify-primop e table))
              ((testop-p e) (simplify-testop e table))
              (t e))))
     (cond
      ((eq e ne) (return (values e env)))
      ((bind-p ne)
       (do ((vars (bind-vars ne) (rest vars))
            (args (sublis env (bind-args ne)) (rest args)))
           ((or (null vars) (null args)))
           (push (cons (first vars) (first args)) env))
       (setf e (bind-body ne)))
      (t (setf e ne))))))

(defun simplify-one-up (e table)
  (loop
   (let ((ne (cond
              ((call-p e) (simplify-call e table))
              ((primop-p e) (simplify-primop e table))
              ((testop-p e) (simplify-testop e table))
              (t e))))
     (if (eq e ne) (return e) (setf e ne)))))

(defun subvars (var env table)
  (if (consp var)
      (mapcar #'(lambda (x) (subvars x env table)) var)
    (let ((val (cdr (assoc var env))))
      (cond
       (val (when (var-p val)
                  (incf (varinfo-count (lookup-varinfo var table))))
            val)
       (t var)))))

(defparameter *generic-call-simplifiers* nil)

(defmacro define-generic-call-simplifier (args &rest body)
  `(progn (push #'(lambda ,args ,@body) *generic-call-simplifiers*)
          nil))

(defparameter *call-simplifiers* (make-hash-table))

(defun get-call-simplifiers (fun) (gethash fun *call-simplifiers*))

(defmacro define-call-simplifier (name args &rest body)
  `(progn (push #'(lambda ,args ,@body) (gethash ',name *call-simplifiers*))
          ',name))

(defun simplify-call (e table)
  (let ((fun (call-fun e)))
    (if (symbolp fun)
        (dolist (s (get-call-simplifiers fun) e)
          (let ((ne (funcall s e table)))
            (unless (eq ne e) (return ne))))
      (dolist (s *generic-call-simplifiers* e)
        (let ((ne (funcall s e table)))
          (unless (eq ne e) (return ne)))))))

(defun simplify-primop (e table)
  (let* ((fun (primop-fun e))
         (args (primop-args e))
         (vars (primop-vars e))
         (form (primop-form e)))
    (if (and (not (member fun '(values values-list)))
             (primop-p form) (eq (primop-fun form) 'values)
             (null (primop-form form))
             (let ((pargs (primop-args form)))
               (and (= (length vars) (length pargs))
                    (every #'eq pargs vars))))

        (remake-cexpr e fun args nil nil)
      (dolist (s (get-primop-simplifiers fun) e)
        (let ((ne (funcall s e table)))
          (unless (eq e ne) (return ne)))))))

(defparameter *primop-simplifiers* (make-hash-table))

(defun get-primop-simplifiers (fun) (gethash fun *primop-simplifiers*))

(defmacro define-primop-simplifier (name args &rest body)
  `(progn (push #'(lambda ,args ,@body) (gethash ',name *primop-simplifiers*))
          ',name))

(defun simplify-testop (e table)
  (dolist (s (get-testop-simplifiers (testop-fun e)) e)
    (let ((ne (funcall s e table)))
      (unless (eq e ne) (return ne)))))

(defparameter *testop-simplifiers* (make-hash-table))

(defun get-testop-simplifiers (fun) (gethash fun *testop-simplifiers*))

(defmacro define-testop-simplifier (name args &rest body)
  `(progn (push #'(lambda ,args ,@body) (gethash ',name *testop-simplifiers*))
          ',name))



;;;
;;; Some Call Simplifiers
;;;

(define-call-simplifier funcall (e table)
  (let* ((args (call-args e))
         (vars (call-vars e))
         (conts (call-conts e))
         (ne (remake-cexpr e (first args) (rest args) vars conts)))
    (set-varinfo-defs vars table ne)
    ne))

(defun set-varinfo-defs (vars table e)
  (dolist (v vars) (setf (varinfo-def (lookup-varinfo v table)) e)))

(define-generic-call-simplifier (e table)
  (let ((entry (lookup-varinfo (call-fun e) table)))
    (if entry
        (let ((def (varinfo-def entry)))
          (if (and (primop-p def) (eq (primop-fun def) 'symbol-function))
              (let* ((vars (call-vars e))
                     (args (call-args e))
                     (conts (call-conts e))
                     (ne (remake-cexpr e (primop-data def) args vars conts)))
                (set-varinfo-defs vars table ne)
                (decf (varinfo-count entry))
                ne)
            e))
      e)))

(define-generic-call-simplifier (e table)
  (let ((fun (call-fun e)))
    (if (and (const-p fun) (symbolp (const-value fun)))
        (let* ((nfun (const-value fun))
               (args (call-args e))
               (vars (call-vars e))
               (conts (call-conts e))
               (ne (remake-cexpr e nfun args vars conts)))
          (set-varinfo-defs vars table ne)
          ne)
      e)))



;;;
;;; Converting Calls to Primops
;;;

(defun call-to-primop (e table template fun)
  (if (or (call-is-mvcall e) (not (args-match-p (call-args e) template)))
      e
    (let* ((args (call-args e))
           (vars (call-vars e))
           (form (call-form e))
           (vi (call-val-ignored e))
           (ctx (call-context e))
           (nvars (cond
                   (vars vars)
                   ((null form) nil)
                   (t (let ((v (make-tmp-var)))
                        (add-varinfo v e table)
                        (list v)))))
           (nform (cond
                   ((or (null form) vars vi) form)
                   (t (let ((entry (lookup-varinfo (first nvars) table)))
                        (incf (varinfo-count entry))
                        (make-primop 'values nvars nil form nil ctx)))))
           (ne (make-primop fun args nvars nform nil ctx)))
      (set-varinfo-defs nvars table ne)
      ne)))

(defun args-match-p (args template)
  (cond
   ((eq template t) t)
   ((numberp template) (= (length args) template))
   ((consp template) (member (length args) template))
   (t nil)))

(defmacro define-primop-call-simplifier (name template &optional (fun name))
  `(define-call-simplifier ,name (e table)
     (call-to-primop e table ',template ',fun)))

(define-primop-call-simplifier + 2)
(define-primop-call-simplifier - 2)
(define-primop-call-simplifier * 2)
(define-primop-call-simplifier / 2)

(define-primop-call-simplifier - 1 uminus)
(define-primop-call-simplifier / 1 udiv)

(define-primop-call-simplifier 1+ 1)
(define-primop-call-simplifier 1- 1)

(define-primop-call-simplifier min 2)
(define-primop-call-simplifier max 2)

(define-primop-call-simplifier <  2)
(define-primop-call-simplifier <= 2)
(define-primop-call-simplifier =  2)
(define-primop-call-simplifier /= 2)
(define-primop-call-simplifier >= 2)
(define-primop-call-simplifier >  2)

(define-primop-call-simplifier cons 2)

(define-primop-call-simplifier car 1)
(define-primop-call-simplifier first 1 car)
(define-primop-call-simplifier cdr 1)
(define-primop-call-simplifier rest 1 cdr)

(define-primop-call-simplifier caar 1)
(define-primop-call-simplifier cadr 1)
(define-primop-call-simplifier cdar 1)
(define-primop-call-simplifier cddr 1)
(define-primop-call-simplifier second 1 cadr)

(define-primop-call-simplifier caaar 1)
(define-primop-call-simplifier caadr 1)
(define-primop-call-simplifier cadar 1)
(define-primop-call-simplifier caddr 1)
(define-primop-call-simplifier cdaar 1)
(define-primop-call-simplifier cdadr 1)
(define-primop-call-simplifier cddar 1)
(define-primop-call-simplifier cdddr 1)
(define-primop-call-simplifier third 1 caddr)

(define-primop-call-simplifier xlisp::%set-car 2)
(define-primop-call-simplifier xlisp::%set-cdr 2)
(define-primop-call-simplifier rplaca 2)
(define-primop-call-simplifier rplacd 2)

(define-primop-call-simplifier xlisp::%set-nth 3)
(define-primop-call-simplifier nth 2)

(define-primop-call-simplifier not 1)
(define-primop-call-simplifier null 1 not)
(define-primop-call-simplifier consp 1)
(define-primop-call-simplifier endp 1)

(define-primop-call-simplifier eq 2)
(define-primop-call-simplifier eql 2)
(define-primop-call-simplifier equal 2)

(define-primop-call-simplifier aref (2 3))
(define-primop-call-simplifier svref 2)
(define-primop-call-simplifier elt 2)
(define-primop-call-simplifier row-major-aref 2)

(define-primop-call-simplifier xlisp::%set-aref (3 4))
(define-primop-call-simplifier xlisp::%set-svref 3)
(define-primop-call-simplifier xlisp::%set-elt 3)
(define-primop-call-simplifier xlisp::%set-rm-aref 3)

(define-call-simplifier values (e table)
  (let ((args (call-args e))
        (vars (call-vars e))
        (form (call-form e))
        (ctx (call-context e)))
    (cond
     ((call-is-mvcall e) e)
     ((null form)
      (make-primop 'values args nil nil nil ctx))
     (vars
      (let* ((nargs (list (if args (first args) (make-const nil)))))
        (make-bind vars nargs form ctx)))
     ((call-val-ignored e)
      (dolist (a args)
        (let ((entry (lookup-varinfo a table)))
          (when entry (decf (varinfo-count entry)))))
      form)
     (t (make-primop 'values args nil form nil ctx)))))

(define-call-simplifier values-list (e table)
  (let ((args (call-args e))
        (vars (call-vars e))
        (form (call-form e))
        (ctx (call-context e)))
    (if (or (call-is-mvcall e) (/= (length args) 1))
        e
      (cond
       ((null form)
        (make-primop 'values-list args nil nil nil ctx))
       (vars
        (let ((ne (make-primop 'car args vars form nil ctx)))
          (setf (varinfo-def (lookup-varinfo (first vars) table)) ne)
          ne))
       ((call-val-ignored e)
        (make-primop 'values-list args nil form nil ctx))
       (t (make-primop 'values-list args nil form nil ctx))))))

(define-call-simplifier xlisp::%struct-ref (e table)
  (let* ((args (call-args e))
         (slot (second args)))
    (if (or (call-is-mvcall e)
            (/= (length args) 2)
            (not (and (const-p slot) (typep (const-value slot) 'fixnum))))
        e
      (let ((ne (call-to-primop e table t 'struct-ref)))
        (setf (primop-data ne) (const-value slot))
        (setf (primop-args ne) (list (first args)))
        ne))))

(define-call-simplifier xlisp::%struct-set (e table)
  (let* ((args (call-args e))
         (slot (second args)))
    (if (or (call-is-mvcall e)
            (/= (length args) 3)
            (not (and (const-p slot) (typep (const-value slot) 'fixnum))))
        e
      (let ((ne (call-to-primop e table t 'set-struct-ref)))
        (setf (primop-data ne) (const-value slot))
        (setf (primop-args ne) (list (first args) (third args)))
        ne))))

(define-call-simplifier xlisp::%struct-type-p (e table)
  (let* ((args (call-args e))
         (type (first args)))
    (if (or (call-is-mvcall e)
            (/= (length args) 2)
            (not (and (const-p type) (symbolp (const-value type)))))
        e
      (let ((ne (call-to-primop e table t 'struct-type-p)))
        (setf (primop-data ne) (const-value type))
        (setf (primop-args ne) (rest args))
        ne))))

(define-primop-call-simplifier get 2)
(define-primop-call-simplifier xlisp::%set-get 3)

(define-primop-call-simplifier slot-value (1 2))



;;;
;;; Simplifiers For Some Primops
;;;

(defun constant-fold-primop (e &optional pred  (fun (primop-fun e)))
  (let ((args (primop-args e)))
    (if (every #'const-p args)
        (let ((vals (mapcar #'const-value args)))
          (if (or (null pred) (every pred vals))
              (wrap-primop-value (make-const (apply fun vals)) e nil)
            e))
      e)))

(defun wrap-primop-value (val e table)
  (let ((vars (primop-vars e))
        (form (primop-form e))
        (ctx (primop-context e)))
    (cond
     (vars (make-bind vars (list val) (primop-form e) ctx))
     (form form)
     (t (when (var-p val) (incf (varinfo-count (lookup-varinfo val table))))
        (make-primop 'values (list val) nil nil nil ctx)))))

(defmacro define-arith-primop-folder (name)
  `(define-primop-simplifier ,name (e table)
     (declare (ignore table))
     (constant-fold-primop e #'numberp)))

(defmacro define-primop-folder (name)
  `(define-primop-simplifier ,name (e table)
     (declare (ignore table))
     (constant-fold-primop e)))

(define-arith-primop-folder +)
(define-arith-primop-folder -)
(define-arith-primop-folder *)
(define-arith-primop-folder /)

(define-arith-primop-folder min)
(define-arith-primop-folder max)

(define-arith-primop-folder 1+)
(define-arith-primop-folder 1-)

(define-arith-primop-folder <)
(define-arith-primop-folder <=)
(define-arith-primop-folder =)
(define-arith-primop-folder /=)
(define-arith-primop-folder >=)
(define-arith-primop-folder >)

(define-primop-folder car)
(define-primop-folder cdr)

(define-primop-folder caar)
(define-primop-folder cadr)
(define-primop-folder cdar)
(define-primop-folder cddr)

(define-primop-folder caaar)
(define-primop-folder caadr)
(define-primop-folder cadar)
(define-primop-folder caddr)
(define-primop-folder cdaar)
(define-primop-folder cdadr)
(define-primop-folder cddar)
(define-primop-folder cdddr)

(define-primop-simplifier + (e table)
  (let* ((args (primop-args e))
         (x (first args))
         (y (second args)))
    (cond
     ((and (const-p x) (eql (const-value x) 1))
      (let* ((vars (primop-vars e))
             (conts (primop-conts e))
             (ne (remake-cexpr e '1+ (list y) vars conts)))
        (set-varinfo-defs vars table ne)
        ne))
     ((and (const-p y) (eql (const-value y) 1))
      (let* ((vars (primop-vars e))
             (conts (primop-conts e))
             (ne (remake-cexpr e '1+ (list x) vars conts)))
        (set-varinfo-defs vars table ne)
        ne))
     (t e))))

(define-primop-simplifier - (e table)
  (let* ((args (primop-args e))
         (x (first args))
         (y (second args)))
    (if (and (const-p y) (eql (const-value y) 1))
        (let* ((vars (primop-vars e))
               (conts (primop-conts e))
               (ne (remake-cexpr e '1- (list x) vars conts)))
          (set-varinfo-defs vars table ne)
          ne)
      e)))

(defun find-car-once-used (var table)
  (if (var-p var)
      (let* ((entry (lookup-varinfo var table))
             (def (varinfo-def entry)))
        (if (and (= (varinfo-count entry) 1)
                 (primop-p def)
                 (eq (primop-fun def) 'cons))
            (first (primop-args def))
          nil))
    nil))

(defun find-cdr-once-used (var table)
  (if (var-p var)
      (let* ((entry (lookup-varinfo var table))
             (def (varinfo-def entry)))
        (if (and (= (varinfo-count entry) 1)
                 (primop-p def)
                 (eq (primop-fun def) 'cons))
            (second (primop-args def))
          nil))
    nil))

(define-primop-simplifier car (e table)
  (let* ((var (first (primop-args e)))
         (car (find-car-once-used var table)))
    (cond
     (car (decf (varinfo-count (lookup-varinfo var table)))
          (wrap-primop-value car e table))
     (t e))))

(define-primop-simplifier cdr (e table)
  (let* ((var (first (primop-args e)))
         (cdr (find-cdr-once-used var table)))
    (cond
     (cdr (decf (varinfo-count (lookup-varinfo var table)))
          (wrap-primop-value cdr e table))
     (t e))))

(defun simplify-cons-cxxr (e table new find)
  (let* ((var (first (primop-args e)))
         (cxr (funcall find var table)))
    (if cxr
        (let ((form (primop-form e))
              (vars (primop-vars e))
              (ctx (primop-context e)))
          (decf (varinfo-count (lookup-varinfo var table)))
          (when (var-p cxr) (incf (varinfo-count (lookup-varinfo cxr table))))
          (make-primop new (list cxr) vars form nil ctx))
      e)))

(defmacro define-cxxr-simplifier (name)
  (let* ((str (symbol-name name))
         (n (length str))
         (c (char str (- n 2)))
         (find (if (eql c #\A) 'find-car-once-used 'find-cdr-once-used))
         (new (intern (concatenate 'string (subseq str 0 (- n 2)) "R"))))
    `(define-primop-simplifier ,name (e table)
       (simplify-cons-cxxr e table ',new #',find))))

(define-cxxr-simplifier caar)
(define-cxxr-simplifier cdar)
(define-cxxr-simplifier cadr)
(define-cxxr-simplifier cddr)

(define-cxxr-simplifier caaar)
(define-cxxr-simplifier cdaar)
(define-cxxr-simplifier cadar)
(define-cxxr-simplifier cddar)
(define-cxxr-simplifier caadr)
(define-cxxr-simplifier cdadr)
(define-cxxr-simplifier caddr)
(define-cxxr-simplifier cdddr)



;;;
;;; Simplifiers For Values Primops
;;;

(define-primop-simplifier values (e table)
  (let ((form (primop-form e)))
    (if (and (call-p form) (call-is-mvcall form))
        (let* ((fun (call-fun form))
               (args (primop-args e))
               (vars (call-vars form))
               (conts (call-conts form))
               (ne (remake-cexpr form fun args vars conts)))
          (setf (call-is-mvcall ne) nil)
          (set-varinfo-defs vars table ne)
          ne)
      e)))

(define-primop-simplifier values (e table)
  (let ((form (primop-form e)))
     (if (or (primop-eq form 'values)
             (primop-eq form 'values-list)
             (and (call-p form)
                  (not (call-is-mvcall form))
                  (symbolp (call-fun form))))
         (dolist (a (primop-args e) form)
           (let ((entry (lookup-varinfo a table)))
             (when entry (decf (varinfo-count entry)))))
       e)))

(define-primop-simplifier values-list (e table)
  (let ((form (primop-form e)))
    (if (and (call-p form)
             (call-is-mvcall form)
             (eq (call-fun form) 'values))
        (let* ((args (primop-args e))
               (vars (call-vars form))
               (conts (call-conts form))
               (nform (if conts (first conts) nil))
               (ctx (call-context form))
               (ne (make-primop 'values-list args vars nform nil ctx)))
          (set-varinfo-defs vars table ne)
          ne)
      e)))



;;;
;;; Simplifiers For Some Testops
;;;

(define-testop-simplifier t (e table)
  (let* ((arg (first (testop-args e)))
         (entry (lookup-varinfo arg table)))
    (if entry
        (let ((def (varinfo-def entry)))
          (if (primop-eq def 'not)
              (let ((nargs (primop-args def))
                    (nconts (reverse (testop-conts e))))
                (decf (varinfo-count entry))
                (dolist (a nargs)
                  (when (var-p a)
                        (let ((entry (lookup-varinfo a table)))
                          (incf (varinfo-count entry)))))
                (remake-cexpr e t nargs nil nconts))
            e))
      e)))

(define-testop-simplifier t (e table)
  (let* ((arg (first (testop-args e)))
         (entry (lookup-varinfo arg table)))
    (if (and entry (= (varinfo-count entry) 1))
        (let ((def (varinfo-def entry)))
          (if (and (primop-p def)
                   (member (primop-fun def)
                           '(consp eq eql equal #|endp < <= = /= >= >|#)))
              (let ((nargs (primop-args def))
                    (conts (testop-conts e)))
                (decf (varinfo-count entry))
                (dolist (a nargs)
                  (when (var-p a)
                        (let ((entry (lookup-varinfo a table)))
                          (incf (varinfo-count entry)))))
                (remake-cexpr e (primop-fun def) nargs nil conts))
            e))
      e)))

(defun constant-fold-testop (e &optional pred  (fun (testop-fun e)))
  (let ((args (testop-args e)))
    (if (every #'const-p args)
        (let ((vals (mapcar #'const-value args)))
          (if (or (null pred) (every pred vals))
              (let ((val (apply fun vals))
                    (conts (testop-conts e)))
                (if val (first conts) (second conts)))
            e))
      e)))
(define-testop-simplifier t (e table)
  (declare (ignore table))
  (constant-fold-testop e nil #'identity))
(define-testop-simplifier consp (e table)
  (declare (ignore table))
  (constant-fold-testop e))
(define-testop-simplifier endp (e table)
  (declare (ignore table))
  (constant-fold-testop e))
(define-testop-simplifier eq (e table)
  (declare (ignore table))
  (constant-fold-testop e))
(define-testop-simplifier eql (e table)
  (declare (ignore table))
  (constant-fold-testop e))
(define-testop-simplifier equal (e table)
  (declare (ignore table))
  (constant-fold-testop e))

(define-testop-simplifier case (e table)
  (declare (ignore table))
  (let ((arg (first (testop-args e))))
    (if (const-p arg)
        (let* ((data (testop-data e))
               (val (const-value arg))
               (i (dotimes (i (length data) i)
                    (let ((entry (aref data i)))
                      (cond
                       ((consp entry) (when (member val entry) (return i)))
                       ((eql val entry) (return i)))))))
          (nth i (testop-conts e)))
      e)))



;;;
;;; Dead Variable Elimination
;;;

(defmacro define-drop-dead-primop-simplifier (name)
  `(define-primop-simplifier ,name (e table)
     (flet ((unused (v) (= (varinfo-count (lookup-varinfo v table)) 0)))
       (let ((vars (primop-vars e))
             (form (primop-form e)))
         (cond
          ((or (and vars (every #'unused vars)) (and (null vars) form))
           (dolist (a (primop-args e))
             (when (var-p a)
                   (decf (varinfo-count (lookup-varinfo a table)))))
           (primop-form e))
          (t e))))))

(define-drop-dead-primop-simplifier cons)

(define-drop-dead-primop-simplifier not)
(define-drop-dead-primop-simplifier consp)
(define-drop-dead-primop-simplifier eq)
(define-drop-dead-primop-simplifier eql)
(define-drop-dead-primop-simplifier equal)

(define-drop-dead-primop-simplifier symbol-function)


;;;;
;;;; Inline Expansion
;;;;

(defun inline-expand (e)
  (let ((ne (inline-expand-built-ins (split-local-functions e))))
    (substitute-once-called-functions (inline-expand-non-recursives ne nil))))



;;;
;;; Expanding Small Non-Recursive Functions
;;;

(defun inline-expand-non-recursives (e env)
  (cond
   ((letrec-p e) (inline-expand-non-recursives-letrec e env))
   ((call-p e) (inline-expand-non-recursives-call e env nil))
   (t (inline-expand-non-recursives-cexpr e env))))

(defun inline-expand-non-recursives-letrec (e env)
  (let* ((nargs (mapcar #'(lambda (x) (inline-expand-non-recursives x env))
                        (letrec-args e)))
         (vars (letrec-vars e))
         (nenv (make-inline-nonrec-env vars nargs env))
         (body (letrec-body e))
         (nbody (if (call-p body)
                    (inline-expand-non-recursives-call body nenv vars)
                  (inline-expand-non-recursives body nenv))))
    (remake-cexpr e nil nargs vars (list nbody))))

(defun make-inline-nonrec-env (vars args env)
  (mapc #'(lambda (v a)
            (let ((body (func-body a)))
              (when (and (null (cexpr-conts body))
                         (not (member (cexpr-fun body) vars))
                         (null (intersection vars (cexpr-args body))))
                    (push (cons v a) env))))
        vars args)
  env)

(defun inline-expand-non-recursives-call (e env skip)
  (let ((def (cdr (assoc (call-fun e) env))))
    (if (or (null def) (member (call-fun e) skip))
        (inline-expand-non-recursives-cexpr e env)
      (let* ((var (make-tmp-var))
             (ndef (alpha-copy def))
             (args (call-args e))
             (vars (call-vars e))
             (form (call-form e))
             (nform (if form (inline-expand-non-recursives form env) nil))
             (ne (remake-cexpr e var args vars (if form (list nform) nil))))
        (make-letrec (list var) (list ndef) ne (call-context e))))))

(defun inline-expand-non-recursives-cexpr (e env)
  (let ((nconts (mapcar #'(lambda (x) (inline-expand-non-recursives x env))
                        (cexpr-conts e))))
    (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts)))



;;;
;;; Expanding Some Built-In Functions
;;;

(defun inline-expand-built-ins (e)
  (cond
   ((letrec-p e) (inline-expand-built-ins-letrec e))
   ((call-p e) (inline-expand-built-ins-call e))
   (t (inline-expand-built-ins-cexpr e))))

(defun inline-expand-built-ins-letrec (e)
  (let ((nargs (mapcar #'inline-expand-built-ins (letrec-args e)))
        (nbody (inline-expand-built-ins (letrec-body e))))
    (remake-cexpr e nil nargs (cexpr-vars e) (list nbody))))

(defun inline-expand-built-ins-call (e)
  (let ((fun (call-fun e)))
    (if (symbolp fun)
        (let ((def (get-global-inline (call-fun e) (length (call-args e)))))
          (if def
              (let* ((var (make-tmp-var))
                     (args (call-args e))
                     (vars (call-vars e))
                     (form (call-form e))
                     (nconts (if form
                                 (list (inline-expand-built-ins form))
                               nil))
                     (ncall (remake-cexpr e var args vars nconts)))
                (make-letrec (list var) (list def) ncall (call-context e)))
            (inline-expand-built-ins-cexpr e)))
      (inline-expand-built-ins-cexpr e))))

(defun inline-expand-built-ins-cexpr (e)
  (let ((nconts (mapcar #'inline-expand-built-ins (cexpr-conts e))))
    (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e) nconts)))

(defparameter *inline-table* (make-hash-table))

(defun set-inline-definition (name argc fun)
  (let ((entry (assoc argc (gethash name *inline-table*))))
    (if entry
        (setf (cdr entry) fun)
      (push (cons argc fun) (gethash name *inline-table*)))
    name))

(defmacro define-inline (name args &rest body)
  (when (intersection args lambda-list-keywords)
        (error "only required arguments allowed in inline definitions"))
  `(set-inline-definition ',name ,(length args) '(lambda ,args ,@body)))

(defun get-global-inline (name argc)
  (let ((def (cdr (assoc argc (gethash name *inline-table*)))))
    (cond
     ((null def) nil)
     ((cexpr-p def) (alpha-copy def))
     (t (let ((cdef (optimize (compile-front-lambda def name nil nil))))
          (set-inline-definition name argc cdef)
          (alpha-copy cdef))))))

(define-inline mapcar (f x)
  (labels ((loop (x v)
             (let ((nv (cons (funcall f (car x)) v))
                   (nx (cdr x)))
               (if (endp nx) (nreverse nv) (loop nx nv)))))
    (if (endp x) nil (loop x nil))))

(define-inline mapcar (f x y)
  (labels ((loop (x y v)
             (let ((nv (cons (funcall f (car x) (car y)) v))
                   (nx (cdr x))
                   (ny (cdr y)))
               (if (and (endp nx) (endp ny)) (nreverse nv) (loop nx ny nv)))))
    (if (or (endp x) (endp y)) nil (loop x y nil))))

(define-inline mapc (f x)
  (labels ((loop (x)
             (let ((nx (cdr x)))
               (funcall f (car x))
               (unless (endp nx) (loop nx)))))
    (unless (endp x) (loop x))
    x))

(define-inline mapc (f x y)
  (labels ((loop (x y)
             (let ((nx (cdr x))
                   (ny (cdr y)))
               (funcall f (car x) (car y))
               (unless (or (endp nx) (endp ny)) (loop nx ny)))))
    (unless (or (endp x) (endp y)) (loop x y))
    x))


;;;;
;;;; Pushing Continuations Into Letrec Expressions
;;;;

(defun push-letrec-conts (e)
  (let ((vars (cexpr-vars e))
        (conts (mapcar #'push-letrec-conts (cexpr-conts e))))
    (if(letrec-p e)
        (let ((vars (letrec-vars e))
              (args (mapcar #'push-letrec-conts (letrec-args e)))
              (body (first conts)))
          (multiple-value-bind (kv kf entries)
                               (find-common-cont body vars)
            (if (and kf
                     (let ((efree nil)
                           (kfree nil))
                       (flet ((register (x free)
                                        (cond
                                         ((eq x e) (setf efree free))
                                         ((eq x kf) (setf kfree (remove kv free))))))
                             (find-freevars e #'register)
                             (subsetp kfree efree)))
)
                (let ((cfuns (find-letrec-tail-funs e entries)))
                  (if cfuns
                      (push-cont-into-letrec e kv kf cfuns)
                    (remake-cexpr e nil args vars conts)))
              (remake-cexpr e nil args vars conts))))
      (remake-cexpr e (cexpr-fun e) (cexpr-args e) vars conts))))

(defun alpha-equal (e1 e2 env)
  (labels ((eqarg (x y env)
             (or (eq x y)
                 (and (var-p x) (var-p y) (eq (sublis env x) y))
                 (and (const-p x) (const-p y)
                      (eql (const-value x) (const-value y)))))
           (alpha-equal-lists (el1 el2 env)
             (do ((el1 el1 (rest el1))
                  (el2 el2 (rest el2)))
                 ((or (null el1) (null el2)) (and (null el1) (null el2)))
                 (unless (alpha-equal (first el1) (first el2) env)
                         (return nil))))
           (eqargs (al1 al2 env)
             (do ((al1 al1 (rest al1))
                  (al2 al2 (rest al2)))
                 ((or (null al1) (null al2)) (and (null al1) (null al2)))
                 (unless (eqarg (first al1) (first al2) env) (return nil)))))
    (and (eq (type-of e1) (type-of e2))
         (eqarg (cexpr-fun e1) (cexpr-fun e2) env)
         (let* ((args1 (cexpr-args e1))
                (args2 (cexpr-args e2))
                (vars1 (cexpr-vars e1))
                (vars2 (cexpr-vars e2))
                (nenv (append (mapcar #'cons vars1 vars2) env)))
           (and (if (letrec-p e1)
                    (alpha-equal-lists args1 args2 nenv)
                  (eqargs args1 args2 env))
                (or (not (primop-p e1))
                    (eql (primop-data e1) (primop-data e2)))
                (or (not (call-p e1))
                    (and (eql (call-val-ignored e1) (call-val-ignored e2))
                         (eql (call-is-mvcall e1) (call-is-mvcall e2))))
                (or (not (func-p e1))
                    (eql (func-type e1) (func-type e2)))
                (alpha-equal-lists (cexpr-conts e1) (cexpr-conts e2) nenv))))))

(defun find-freevars (e record)
  (flet ((free-in-list (list)
           (let ((free nil))
             (dolist (e list free)
               (setf free (append free (find-freevars e record)))))))
    (let ((cfree (free-in-list (cexpr-conts e)))
          (afree (let ((args (cexpr-args e)))
                   (if (letrec-p e)
                       (free-in-list args)
                     (let ((afree (remove-if-not #'var-p args)))
                       (if (call-p e)
                           (let ((fun (call-fun e)))
                             (if (var-p fun)
                                 (cons fun afree)
                               afree))
                         afree)))))
          (vars (cexpr-vars e)))
      (let ((free (set-difference (union cfree afree) vars)))
        (when record (funcall record e free))
        free))))

(defun find-common-cont (e funs)
  (let ((stack (list e))
        (kv nil)
        (kf nil)
        (starts nil))
    (loop
     (if (null stack)
         (return (values kv kf starts))
       (let* ((next (pop stack))
              (args (cexpr-args next)))
         (when (if (letrec-p next)
                   (some #'(lambda (x) (any-uses-p x funs)) args)
                 (some #'(lambda (x) (member x funs)) args))
               (return nil))
         (when (and (call-p next) (member (cexpr-fun next) funs))
               (let ((form (call-form next)))
                 (cond
                  ((null form) (return nil))
                  ((null kf) (let ((vars (call-vars next))
                                   (vi (call-val-ignored next)))
                               (setf kv (if vars (first vars) (if vi nil t)))
                               (setf kf form)
                               (pushnew (call-fun next) starts)))
                  (t (let* ((vars (call-vars next))
                            (vi (call-val-ignored next))
                            (nkv (if vars (first vars) (if vi nil t)))
                            (nkf form))
                       (if (and (var-p kv) (var-p nkv))
                           (unless (alpha-equal kf nkf (list (cons kv nkv)))
                                   (return nil))
                         (unless (and (eq kv nkv) (alpha-equal kf nkf nil))
                                 (return nil)))
                       (pushnew (call-fun next) starts))))))
         (setf stack (append (cexpr-conts next) stack)))))))

(defun any-uses-p (e funs)
  (flet ((any-uses (x) (any-uses-p x funs))
         (is-one (x) (member x funs)))
    (or (is-one (cexpr-fun e))
        (if (letrec-p e)
            (some #'any-uses (letrec-args e))
          (some #'is-one (cexpr-args e)))
        (some #'any-uses (cexpr-conts e)))))

(defstruct (plc-entry (:type list) (:constructor make-plc-entry (name)))
  name conts tail-callees)
(defstruct plc-table entries (hash (make-hash-table)))

(defun add-plc-entry (name table)
  (let ((entry (make-plc-entry name)))
    (push entry (plc-table-entries table))
    (setf (gethash name (plc-table-hash table)) entry)))

(defun lookup-plc-entry (name table)
  (gethash name (plc-table-hash table)))

(defun build-plc-table (e)
  (let ((table (make-plc-table))
        (args (letrec-args e))
        (vars (letrec-vars e)))
    (dolist (v vars) (add-plc-entry v table))
    (init-plc-table-letrec-funs vars args table)
    table))

(defun init-plc-table (e entry table)
  (cond
   ((letrec-p e) (init-plc-table-letrec e entry table))
   ((call-p e) (init-plc-table-call e entry table))
   (t (init-plc-table-cexpr e entry table))))

(defun init-plc-table-letrec (e entry table)
  (let ((args (letrec-args e))
        (vars (letrec-vars e)))
    (dolist (v vars) (add-plc-entry v table))
    (init-plc-table-letrec-funs vars args table)
    (init-plc-table (letrec-body e) entry table)))

(defun init-plc-table-letrec-funs (vars args table)
  (do ((vars vars (rest vars))
       (args args (rest args)))
      ((or (null args) (null vars)))
      (let ((entry (lookup-plc-entry (first vars) table)))
        (init-plc-table (first args) entry table))))

(defun init-plc-table-call (e entry table)
  (init-plc-table-cexpr-args (cexpr-args e) table)
  (let* ((fun (call-fun e))
         (form (call-form e))
         (fentry (lookup-plc-entry fun table)))
    (cond
     (form
      (when fentry (pushnew 'unknown (plc-entry-conts entry)))
      (init-plc-table form entry table))
     (t (when fentry (pushnew fun (plc-entry-tail-callees entry)))))))

(defun init-plc-table-cexpr (e entry table)
  (init-plc-table-cexpr-args (cexpr-args e) table)
  (dolist (c (cexpr-conts e))
    (init-plc-table c entry table)))

(defun init-plc-table-cexpr-args (args table)
  (dolist (a args)
    (let ((entry (lookup-plc-entry a table)))
      (when entry (pushnew 'unknown (plc-entry-conts entry))))))

(defun compute-plc-table (e starts)
  (let* ((table (build-plc-table e))
         (entries (plc-table-entries table)))
    (dolist (entry entries)
      (when (member (plc-entry-name entry) starts)
            (pushnew 'known (plc-entry-conts entry))))
    (loop
     (let ((changed nil))
       (dolist (entry entries)
         (let ((conts (plc-entry-conts entry)))
           (dolist (c (plc-entry-tail-callees entry))
             (let* ((centry (lookup-plc-entry c table))
                    (cconts (plc-entry-conts centry)))
               (unless (subsetp conts cconts)
                       (setf changed t)
                       (setf (plc-entry-conts centry)
                             (union conts cconts)))))))
       (unless changed (return))))
    table))

(defun find-letrec-tail-funs (e starts)
  (let ((table (compute-plc-table e starts))
        (funs nil))
    (dolist (entry (plc-table-entries table) funs)
      (let ((conts (plc-entry-conts entry)))
        (when (member 'known conts)
              (if (member 'unknown conts)
                  (return nil)
                (push (plc-entry-name entry) funs)))))))

(defun push-cont-into-letrec (e kv kf funs)
  (let ((kfun (let ((ctx (cexpr-context kf))
                    (kvars (if (var-p kv) (list kv) nil)))
                (make-func nil kvars kf 'lambda ctx)))
        (k (make-tmp-var)))
    (labels ((pushcont (e)
               (if (letrec-p e)
                   (pushcont-letrec e)
                 (let ((conts (cexpr-conts e)))
                   (if conts
                       (remake-cexpr e (cexpr-fun e) (cexpr-args e)
                                     (cexpr-vars e) (mapcar #'pushcont conts))
                     (substitute-continuation e kv k funs)))))
             (pushcont-letrec (e)
               (let* ((args (cexpr-args e))
                      (vars (cexpr-vars e))
                      (nargs (mapcar #'(lambda (v a)
                                         (if (member v funs) (pushcont a) a))
                                     vars args))
                      (nconts (list (pushcont (letrec-body e)))))
                 (remake-cexpr e nil nargs vars nconts))))
      (let ((args (mapcar #'pushcont (letrec-args e)))
            (conts (list (drop-conts-from-letrec-body e funs))))
        (remake-cexpr e nil (cons kfun args) (cons k (cexpr-vars e)) conts)))))

(defun drop-conts-from-letrec-body (e funs)
  (labels ((fix-calls (e)
            (if (and (call-p e) (member (call-fun e) funs))
                (remake-cexpr e (call-fun e) (call-args e) nil nil)
              (let ((nconts (mapcar #'fix-calls (cexpr-conts e))))
                (remake-cexpr e (cexpr-fun e) (cexpr-args e) (cexpr-vars e)
                              nconts)))))
    (fix-calls (letrec-body e))))

