(in-package "NEWCMP")

;;;;
;;;; Compiler Top Level
;;;;


;;;
;;; Public Variables
;;;

(defvar *compile-verbose* t)
(defvar *compile-print* t)
(defvar *compile-file-pathname*)
(defvar *compile-file-truename*)
(export '(*compile-print-symbol-package* *save-lambda-expressions*))
(defvar *compile-print-symbol-package* nil)
(defvar *save-lambda-expressions* nil)


;;;
;;; Private Variables
;;;

(defvar *global-calls*)
(defvar *global-callinfo*)
(defvar *inline-definitions* nil)


;;;
;;; Macros
;;;

(defmacro with-compiler-environment (&rest body)
  `(let ((xlisp::*simplify-setf* nil)
         (xlisp::*cmp-setf* nil)
         (xlisp::*cmp-structs* nil)
         (xlisp::*cmp-macros* nil)   ;;**** remove eventually?
         (xlisp::*cmp-specials* nil) ;;**** remove eventually?
         (xlisp::*global-environment* xlisp::*global-environment*)
         (*global-calls* nil)
         (*global-callinfo* nil)
         (*inline-definitions* nil))
     ,@body))


;;;
;;; Private Functions
;;;

(defun compile-front-form (e env)
  (compile-front-lambda `(lambda () ,e) nil env (make-context e)))

(defun compile-front-lambda (fun name env ctx)
  (substitute-values
   (add-cells
    (convert-lambda name fun (make-info nil env) nil nil ctx))))

(defun optimize (e)
  (dotimes (i 5 e)
    (let ((ne (inline-expand
               (simplify 
                (drop-safe-cells
                 (substitute-once-called-functions
                  (push-letrec-conts
                   (substitute-once-called-functions
                    (drop-safe-tags e)))))))))
      (when (eq e ne) (return e))
      (setf e ne))))

(defun compile-back-1 (e)
  (let ((ee (lambda-lift (lift-constants (split-all-functions e)))))
    (multiple-value-bind (code lits)
                         (gencode ee)
      (let* ((labels (remove-if-not #'var-p code))
             (pairs (mapcar #'(lambda (x) (cons x (gensym (var-name x))))
                            labels))
             (c (sublis pairs code)))
        (list c lits (first c))))))

(defun compile-back (e)
  (xlscmp::assemble (compile-back-1 e)))

(defun pcvt (e)
  (let ((f (if (lambda-expression-p e) e `(lambda () ,e))))
    (with-compiler-environment
     (optimize (compile-front-lambda f nil nil (make-context f))))))

(defun pcmp (e) (compile-back-1 (pcvt e)))

(defun style-warning-p (c) (typep c 'style-warning))

(defun compile-lambda-expression (fun name)
  (with-compiler-environment
   (let* ((macrop (eq (first fun) 'macro))
          (fun (if macrop (cons 'lambda (rest fun)) fun))
          (env xlisp::*global-environment*)
          (cfun (byte-code-close
                 (compile-back
                  (optimize
                   (compile-front-lambda fun name env (make-context fun)))))))
     (if macrop (coerce-to-macro cfun) cfun))))

(defun compile-function (f fun)
  (cond
   ((or (typep fun 'subr) (typep fun 'fsubr) (compiled-function-p fun))
    (values fun nil))
   ((typep fun 'closure)
    (multiple-value-bind (flam env fname)
                         (function-lambda-expression fun)
      (when env (error "COMPILE can only compile top level definitions"))
      (values (compile-lambda-expression flam (if f f fname)) flam)))
   ((and (consp fun) (eq (first fun) 'lambda))
    (values (compile-lambda-expression fun f) fun))
   (t (error "bad argument type -- ~s" fun))))

(defun print-compiled-form (form stream sp)
  (let ((*print-readably* t)
        (*print-symbol-package* sp))
    (format stream "~s~%" form)))

(defun accumulate-compiled-form (form outstuff)
  (rplacd (last outstuff) (list form)))

(defun compile-one (e compile-time-too env stream)
  (setq e (expand-one e env))
  (case (first e)
        (progn
          (dolist (e (rest e)) (compile-one e compile-time-too env stream)))
        (defun
         (let* ((name (second e))
                (db (xlisp::split-declarations (rest (rest (rest e)))))
                (decls (first db))
                (b (second db))
                (doc (third db))
                (def (if (or *save-lambda-expressions* (inline-p name env))
                         `(lambda ,(third e) ,@decls ,@b)
                       nil))
                (denv (if def (if (eq env xlisp::*global-environment*) nil t) nil))
                (fexpr (if (not-inline-p name env)
                           `(lambda ,(third e) ,@decls ,@b)
                         `(labels ((,name ,(third e) ,@decls ,@b)) #',name))))
           (register-global-fun name (third e))
           (register-inline-definition name def)
           (let* ((cexpr (compile-back (optimize (compile-front-form fexpr env))))
                  (cform (if doc
                             `(progn
                                (xlisp::%set-get ',name 'function-documentation ,doc)
                                (xlisp::install-function ',name ,cexpr ',def ,denv))
                           `(xlisp::install-function ',name ,cexpr ',def ,denv))))
             (accumulate-compiled-form cform stream))))
        (macrolet
         (flet ((mkmacro (x)
                         (let ((name (first x))
                               (args (second x))
                               (body (rest (rest x))))
                           (list name (parse-macro name args body env)))))
               (multiple-value-bind (decls body)
                                    (split-decls-body (rest (rest e)) nil)
                 (let* ((macs (mapcar #'mkmacro (second e)))
                        (nenv (augment-environment env :macro macs))
                        (nnenv (augment-environment nenv :declare decls)))
                   (compile-one body compile-time-too nnenv stream)))))
        (eval-when
         (let* ((sits (second e))
                (body `(progn ,@(rest (rest e))))
                (lt (or (member 'load sits) (member :load-toplevel sits)))
                (ct (or (member 'compile sits)
                        (member :compile-toplevel sits)))
                (ex (or (member 'eval sits) (member :execute sits))))
           (cond
            ((or (and lt ct)
                 (and lt (not ct) ex compile-time-too))
             (compile-one body t env stream))
            ((or (and lt (not ct) (not compile-time-too))
                 (and lt (not ct) (not ex)))
             (compile-one body nil env stream))
            ((or (and (not lt) ct)
                 (and (not lt) (not ct) ex compile-time-too))
             (eval body)))))
        (locally
         (multiple-value-bind (decls body)
                              (split-decls-body (rest e) nil)
           (let ((nenv (augment-environment env :declare decls)))
             (compile-one body compile-time-too nenv stream))))
        (symbol-macrolet
         (check-specop-argument-count 'symbol-macrolet (rest e) 1 0 t)
         (check-bindings 'symbol-macrolet (second e))
         (multiple-value-bind (decls body)
                              (split-decls-body (rest (rest e)) nil)
           (let* ((nenv (augment-environment env :symbol-macro (second e)))
                  (nnenv (augment-environment nenv :declare decls)))
             (compile-one body compile-time-too nnenv stream))))
        (t (if compile-time-too (eval e))
           (compile-form e env stream))))

(defun expand-one (e env)
  (loop
   (if (member (first e) '(progn macrolet eval-when defun)) (return e))
   (multiple-value-bind (ee flag) (expand-expression-1 e env)
     (if flag
         (setq e ee)
       (return e)))))

(defun compile-form (e env stream)
  (let ((cform (compile-back (optimize (compile-front-form e env)))))
    (accumulate-compiled-form cform stream)))

(defun global-callinfo (fun)
  (if (boundp '*global-callinfo*)
      (let ((entry (assoc fun *global-callinfo*)))
        (cond
         (entry (setf (second entry) t) (third entry))
         (t nil)))
    nil))

(defun register-global-call (fun args ctx)
  (declare (ignore ctx))
  (when (and (boundp '*global-callinfo*) (boundp '*global-calls*))
        (let ((entry (assoc fun *global-callinfo*)))
          (if entry
              (setf (second entry) t)
            (let ((entry (assoc fun *global-calls*)))
              (if entry
                  (pushnew (length args) (rest entry))
                (push (list fun (length args)) *global-calls*)))))))

(defun register-global-fun (fun args)
  (let ((ci (lambda-list-callinfo args))
        (sys-ci (sysfun-callinfo fun))
        (old-ci (global-callinfo fun))
        (used nil))
    (when sys-ci
          (issue-style-warning "redefining built-in function ~s~
                                ~@[ with incompatible arguments~]"
                               fun (not (callinfo-equal ci sys-ci))))
    (when (and old-ci (not (callinfo-equal ci old-ci)))
          (issue-style-warning "redefining function ~s ~
                                with incompatible arguments"
                               fun))
    (when (and (null old-ci) (boundp '*global-calls*))
          (let ((entry (assoc fun *global-calls*)))
            (when entry
                  (dolist (n (rest entry))
                    (when (check-call n ci)
                          (issue-style-warning "definition of ~s is not ~
                                                compatible with earlier calls"
                                               fun)
                          (return)))
                  (setf used t)
                  (setf *global-calls* (remove entry *global-calls*)))))
    (when (boundp '*global-callinfo*)
          (let ((entry (assoc fun *global-callinfo*)))
            (if entry
                (setf (third entry) ci)
              (push (list fun used ci) *global-callinfo*))))))

(defun callinfo-equal (x y)
  (or (equal x y)
      (and (eql (callinfo-nreq x) (callinfo-nreq y))
           (eql (callinfo-nopt x) (callinfo-nopt y))
           (eql (callinfo-restp x) (callinfo-restp y))
           (eql (callinfo-keyp x) (callinfo-keyp y))
           (let ((xk (callinfo-keys x))
                 (yk (callinfo-keys y)))
             (and (subsetp xk yk) (subsetp yk xk)))
           (eql (callinfo-other x) (callinfo-other y)))))

(defun check-function-usage ()
  (let ((unused (mapcar #'first (remove-if #'second *global-callinfo*)))
        (used (mapcar #'first *global-calls*)))
    (when unused
          (format t "~&; Functions defined but not used: ~{~s ~}~%"
                  (remove-duplicates unused)))
    (when used
          (format t "~&; Functions used but not defined: ~{~s ~}~%"
                  (remove-duplicates used)))))

(defun register-inline-definition (name def)
  (push (cons name def) *inline-definitions*))

(defun get-inline-definition (name)
  (let ((entry (assoc name *inline-definitions*)))
    (if entry
        (rest entry)
      (if (fboundp name)
          (let ((fun (symbol-function name)))
            (multiple-value-bind (def env)
                                 (function-lambda-expression fun)
              (if env nil def)))
        nil))))


;;;
;;; Public Functions
;;;

(defun compile (f &optional (fun (symbol-function f)))
  (let ((conditions nil))
    (handler-bind ((condition #'(lambda (c) (push c conditions))))
      (multiple-value-bind (cfun flam)
                           (compile-function f fun)
        (let ((warns (if conditions t nil))
              (errs (not (every #'style-warning-p conditions))))
          (if f
              (let ((def (if (or *save-lambda-expressions* (inline-p f nil))
                             flam
                           nil)))
                (values (xlisp::install-function f cfun def nil) warns errs))
            (values cfun warns errs)))))))

(defun compile-file (file &key
                          (output-file (merge-pathnames ".fsl" file))
                          (temporary-file (merge-pathnames "cmptmp.fsl" file))
                          (load nil)
                          (print *compile-print*)
                          (verbose *compile-verbose*)
                          ((:print-symbol-package sp) ;;**** do this cleaner
                           *compile-print-symbol-package*))
  (let* ((*package* *package*)
         (*readtable* *readtable*)
         (iname (merge-pathnames ".lsp" file))
         (*compile-file-pathname* iname)
         (*compile-file-truename* (truename iname))
         (conditions nil)
         (eof (list 'eof)))
    (handler-bind ((condition #'(lambda (c) (push c conditions))))
      (with-compiler-environment
        (with-open-file (in iname)
          (with-open-file (out temporary-file :direction :output)
            (when verbose
                  (format t "~&; compiling file ~s~%" iname)
                  (force-output))
            (let ((major xlisp::*fsl-major-version*)
                  (minor xlisp::*fsl-minor-version*))
              (print-compiled-form `(xlisp::check-fsl-version ,major ,minor)
                                   out sp))
            (do ((e (read in nil eof) (read in nil eof)))
                ((eq e eof))
                (when print
                      (let ((*print-level* 2)
                            (*print-length* 3))
                        (format t "~&; compiling ~s ... " e)
                        (force-output)))
                (handler-case
                 (let ((outstuff (list 'progn)))
                   (compile-one e nil xlisp::*global-environment* outstuff)
                   (case (length outstuff)
                         (1 nil)
                         (2 (print-compiled-form (second outstuff) out sp))
                         (t (print-compiled-form outstuff out sp))))
                 (error (c)
                        (format *error-output* "~&Compiler error: ~a~%" c)
                        (print-compiled-form e out sp)))
                (when print
                      (format t "done~%")
                      (force-output)))))
        (when verbose (check-function-usage))
        (rename-file temporary-file output-file)
        (when load (load output-file))
        (values (truename output-file)
                (if conditions t nil)
                (not (every #'style-warning-p conditions)))))))

(defun compile-file-pathname (file &key
                                   (output-file (merge-pathnames ".fsl" file))
                                   &allow-other-keys)
  output-file)
